from typing import Union
import math

import numpy as np
import matplotlib.pyplot as plt

import pyco.basis
import pyco.waarde
import pyco.knoop
import pyco.lijn

class pc:
    BasisObject = pyco.basis.BasisObject
    Waarde = pyco.waarde.Waarde
    Knoop = pyco.knoop.Knoop
    Lijn = pyco.lijn.Lijn

    
class VormFuncties:
    """Statische hulpfuncties voor uitrekenen eigenschappen vorm."""

    @classmethod
    def punt_aan_linker_zijde(cls, la, lb, p):
        """Checkt of punt p aan linker kant van lijn la-lb ligt."""
        return (((lb[0] - la[0]) * (p[1] - la[1])
                 - (lb[1] - la[1]) * (p[0] - la[0])) > 0)

    @classmethod
    def punt_aan_rechter_zijde(cls, la, lb, p):
        """Checkt of punt p aan rechter kant van lijn la-lb ligt."""
        return (((lb[0] - la[0]) * (p[1] - la[1])
                 - (lb[1] - la[1]) * (p[0] - la[0])) < 0)

    @classmethod
    def punt_op_lijn(cls, la, lb, p):
        """Checkt of punt p op lijn la-lb ligt, tussen la en lb."""
        op_lijn = (((lb[0] - la[0]) * (p[1] - la[1])
                    - (lb[1] - la[1]) * (p[0] - la[0])) == 0)
        if not op_lijn:
            return False
        if la[0] == lb[0]:     # verticale lijn
            return ((p[1] <= lb[1] and p[1] >= la[1]) or
                    (p[1] >= lb[1] and p[1] <= la[1]))
        else:
            return ((p[0] <= lb[0] and p[0] >= la[0]) or
                    (p[0] >= lb[0] and p[0] <= la[0]))

    @classmethod
    def lijn_raakt_lijn(cls, l1a, l1b, l2a, l2b):
        """Checkt of lijn l1a-l1b de lijn l2a-l2b kruist of raakt, op de
        lijnstukken TUSSEN de punten en INCLUSIEF de punten zelf."""
        if (cls.punt_op_lijn(l1a, l1b, l2a) or
                cls.punt_op_lijn(l1a, l1b, l2b) or
                cls.punt_op_lijn(l2a, l2b, l1a) or
                cls.punt_op_lijn(l2a, l2b, l1b)):
            # uiteinde lijn ligt op andere lijn
            return True

        l1xmin = min(l1a[0], l1b[0])
        l1xmax = max(l1a[0], l1b[0])
        l2xmin = min(l2a[0], l2b[0])
        l2xmax = max(l2a[0], l2b[0])
        l1ymin = min(l1a[1], l1b[1])
        l1ymax = max(l1a[1], l1b[1])
        l2ymin = min(l2a[1], l2b[1])
        l2ymax = max(l2a[1], l2b[1])
        if (l1xmin > l2xmax or l2xmin > l1xmax or
                l1ymin > l2ymax or l2ymin > l1ymax):
            # lijn ligt helemaal links/rechts/onder/boven andere lijn
            return False

        # l1 = a * x + b
        # l2 = c * x + d
        if l1b[0] - l1a[0] == 0:        # l1 is een verticale lijn
            x = l1a[0]
            if l2b[0] - l2a[0] == 0: 	# l2 is ook verticaal
                return False
            else:
                c = 1.0 * (l2b[1] - l2a[1]) / (l2b[0] - l2a[0])
                d = l2a[1] - l2a[0] * c
                y = c * x + d
                if ((y >= l1a[1] and y <= l1b[1]) or
                         (y >= l1b[1] and y <= l1a[1])):
                    return True
                else:
                    return False
        elif l2b[0] - l2a[0] == 0:          # l2 is een verticale lijn
            x = l2a[0]
            if l1b[0] - l1a[0] == 0:	    # l1 is ook verticaal
                return False
            else:
                a = 1.0 * (l1b[1] - l1a[1]) / (l1b[0] - l1a[0])
                b = l1a[1] - l1a[0] * a
                y = a * x + b
                if ((y >= l2a[1] and y <= l2b[1]) or
                         (y >= l2b[1] and y <= l2a[1])):
                    return True
                else:
                    return False
        else:                           # beide lijnen zijn diagonaal
            a = 1.0 * (l1b[1] - l1a[1]) / (l1b[0] - l1a[0])
            b = l1a[1] - l1a[0] * a
            c = 1.0 * (l2b[1] - l2a[1]) / (l2b[0] - l2a[0])
            d = l2a[1] - l2a[0] * c
            # l1 gelijk aan l2
            if a - c == 0:              # parallel
                return False
            x = 1.0 * (d - b) / (a - c)
        if x >= l1xmin and x >= l2xmin and x <= l1xmax and x <= l2xmax:
            return True
        else:
            return False

    @classmethod
    def bereken_hoek(cls, p1, p2, p3, clockw=False):
        """Berekent de hoek in graden tussen punten p1, p2 en p3."""
        det = ((p2[0] - p1[0]) * (p3[1] - p1[1])
                   - (p2[1] - p1[1]) * (p3[0] - p1[0]))
        vec1 = [p2[0] - p1[0], p2[1] - p1[1]]
        vec2 = [p3[0] - p2[0], p3[1] - p2[1]]
        product = vec1[0] * vec2[0] + vec1[1] * vec2[1]

        factor = 0.0
        if det != 0:
            factor = 1.0 * product / det
        length1 = math.sqrt(vec1[0]**2 + vec1[1]**2)
        length2 = math.sqrt(vec2[0]**2 + vec2[1]**2)
        denominator = length1 * length2

        angle = 90.0
        if denominator != 0:
            angle = ((math.acos(1.0 * product / denominator)) *
                (360.0 / (2 * math.pi)))

        # angle is defined counter clockwise
        if angle == 0 or angle == 90 or angle == 180:   # angle is orthogonal
            if det == 0:                         # vec2 in line with vec1
                if angle == 180:
                    angle = 0               # 0
                else:
                    angle = 180             # 180
            else:                                # vec2 is not in line with vec1
                if det < 0:                      # p3 is right of vec1
                    angle = 90              # 90
                else:                            # p3 is left of vec1
                    angle = 270             # 270
        else:                                    # angle is not orthogonal
            if det < 0:                          # p3 is right of vec1
                if factor < 0:                   # p3 is further than vec1
                    angle = 180 - angle     # 90-180
                else:                            # p3 is backwards of vec1
                    angle = 180 - angle     # 0-90
            else:                                # p3 is left of vec1
                if factor > 0:                   # p3 is further than vec1
                    angle = 180 + angle     # 180-270
                else:                            # p3 is backwards of vec1
                    angle = 180 + angle     # 270-360
        if clockw:
            angle = (360 - angle) % 360
        return angle

    @classmethod
    def interpoleer_over_lijn(cls, p1, p2, x):
        """Berekent de y-waarde (bij een gegeven x-waarde) over lijn door p1 en p2. Als er geen waarde bepaald kan worden dan: 99999."""
        if p2[0] - p1[0] == 0 or not (
                isinstance(x, int) or isinstance(x, float)):
            return 99999
        else:
            a = 1.0 * (p2[1] - p1[1]) / (p2[0] - p1[0])
            b = 1.0 * p1[1] - p1[0] * a
            return a * x + b

    @classmethod
    def bereken_oppervlakte_driehoek(cls, p1, p2, p3):
        A = (p1[0] * (p2[1] - p3[1]) + p2[0] * (p3[1] - p1[1]) +
            p3[0] * (p1[1] - p2[1])) / 2.0
        return abs(A)

    @classmethod
    def bereken_zwaartepunt_driehoek(cls, p1, p2, p3):
        ncx = (p1[0] + p2[0] + p3[0]) / 3.0
        ncy = (p1[1] + p2[1] + p3[1]) / 3.0
        return [ncx, ncy]

    @classmethod
    def bereken_traagheidsmomenten_driehoek(cls, p1, p2, p3):
        nc = cls.bereken_zwaartepunt_driehoek(p1, p2, p3)
        ncx = nc[0]
        ncy = nc[1]
        c = [p1, p2, p3]
        Ixx = 0.0
        Iyy = 0.0
        Ixy = 0.0
        l = len(c)
        for i in range(l):
            p = [c[i][0] - ncx, c[i][1] - ncy]
            p_volgende = [c[(i + 1) % l][0] - ncx,
                          c[(i + 1) % l][1] - ncy]
            a = abs(p[0] * p_volgende[1] - p_volgende[0] * p[1])
            Ixx += (p[0] * p[0] + p[0] * p_volgende[0] +
                    p_volgende[0] * p_volgende[0]) * a
            Iyy += (p[1] * p[1] + p[1] * p_volgende[1] +
                    p_volgende[1] * p_volgende[1]) * a
            Ixy += (p[0] * p_volgende[1] + 2 * p[0] * p[1] +
                2 * p_volgende[0] * p_volgende[1] + p_volgende[0] * p[1]) * a
        Ixx /= 12.0
        Iyy /= 12.0
        Ixy /= 24.0
        return [Ixx, Iyy, Ixy]

    @classmethod
    def genereer_net(cls, np_array):
        """Triangulatie van niet-convexe polygoon."""

        coordinaten = []
        coordinaten_links_rechts = []

        # zorgen dat laatste knoop niet zelfde is als eerste
        if np_array[0].tolist() == np_array[-1].tolist():
            np_array = np.delete(np_array, (-1), axis=0)

        # zorgen dat omtrek lijnen elkaar nergens kruisen/raken
        l = len(np_array)
        for i in range(l):
            # kies een lijn
            p1 = [np_array[i][0], np_array[i][1]]
            p2 = [np_array[(i+1)%l][0], np_array[(i+1)%l][1]]
            for ii in range(i, l):
                # check alle andere lijnen
                if i != (ii-1)%l and i != ii and i != (ii+1)%l:
                    pp1 = [np_array[ii][0], np_array[ii][1]]
                    pp2 = [np_array[(ii+1)%l][0], np_array[(ii+1)%l][1]]
                    if cls.lijn_raakt_lijn(p1, p2, pp1, pp2):
                        raise ValueError('Omtrek van vorm mag zichzelf nergens raken/kruisen. Volgende lijnen doen dat wel: Lijn({}, {}) & Lijn({}, {})'.format(p1, p2, pp1, pp2))

        def verwijder_punt_op_lijn(c):
            nonlocal coordinaten_links_rechts
            punt_op_lijn_gevonden = False
            c = c if (isinstance(c, list) or isinstance(c, tuple)) else []
            n = len(c)

            for i in range(n):
                x = c[i][0]
                y = c[i][1]
                vorige_x = c[(i - 1 + n) % n][0]
                vorige_y = c[(i - 1 + n) % n][1]
                volgende_x = c[(i + 1) % n][0]
                volgende_y = c[(i + 1) % n][1]
                if cls.punt_op_lijn([vorige_x, vorige_y],
                                        [volgende_x, volgende_y], [x, y]):
                    del c[i]
                    punt_op_lijn_gevonden = True
                    break
            if punt_op_lijn_gevonden:
                verwijder_punt_op_lijn(c)
            else:
                n = len(c)
                for i in range(n):
                    coordinaten.append([c[i][0], c[i][1], i])
                coordinaten_links_rechts = sorted(
                        coordinaten, key=lambda x: x[0])

        verwijder_punt_op_lijn(np_array.tolist())

        # uitzoeken of vorm met klok mee gaat of tegen klok in
        linker_c = coordinaten_links_rechts[0]
        l = len(coordinaten)
        volgende_c = coordinaten[(linker_c[2] + 1) % l]
        vorige_c = coordinaten[(linker_c[2] - 1 + l) % l]
        volgende_y = volgende_c[1]
        vorige_y = vorige_c[1]
        coordinaten_met_klok_mee = True
        if volgende_c[0] > vorige_c[0]:
            interpol_y = cls.interpoleer_over_lijn(
                    linker_c, volgende_c, vorige_c[0])
            if interpol_y != 99999:
                volgende_y = interpol_y
        else:
            interpol_y = cls.interpoleer_over_lijn(
                    linker_c, vorige_c, volgende_c[0])
            if interpol_y != 99999:
                vorige_y = interpol_y
        if volgende_y < vorige_y:
            coordinaten_met_klok_mee = False

        # hoek berekenen en toevoegen aan array
        for i in range(l):
            p = coordinaten[i]
            vorige_p = coordinaten[(i - 1 + l) % l]
            volgende_p = coordinaten[(i + 1) % l]
            hoek = cls.bereken_hoek(
                    vorige_p, p, volgende_p,
                    not coordinaten_met_klok_mee)
            coordinaten[i].append(hoek)

        np_array = np.array([[x, y, h] for x, y, _, h in coordinaten])

        driehoeken = []
        driehoeken_lijnen_intern = []

        aantal_iteraties = 0

        def driehoek_iterator(polys):
            nonlocal driehoeken
            nonlocal driehoeken_lijnen_intern
            nonlocal aantal_iteraties

            aantal_iteraties += 1
            if aantal_iteraties > 9999:
                raise RecursionError('maximale aantal iteraties bereikt')

            gevonden = False
            for poly_i, poly in enumerate(polys):
                if gevonden:
                    break
                l = len(poly)
                if l > 3:
                    for i in range(l):
                        p = [poly[i][0], poly[i][1],
                            poly[i][2], poly[i][3]] # x,y,i,angle
                        p_vorige = [poly[(i-1+l)%l][0], poly[(i-1+l)%l][1],
                              poly[(i-1+l)%l][2], poly[(i-1+l)%l][3]]
                        p_volgende = [poly[(i+1)%l][0], poly[(i+1)%l][1],
                            poly[(i+1)%l][2], poly[(i+1)%l][3]]
                        over_lijn = False

                        for i_sub in range(l):
                            p_sub = [poly[i_sub][0], poly[i_sub][1],
                                poly[i_sub][2], poly[i_sub][3]]
                            p_sub_volgende = [poly[(i_sub+1)%l][0],
                                poly[(i_sub+1)%l][1],
                                poly[(i_sub+1)%l][2],
                                poly[(i_sub+1)%l][3]]
                            if (not (p_vorige[0] == p_sub[0]
                                    and p_vorige[1] == p_sub[1])
                                and not (p_vorige[0] == p_sub_volgende[0]
                                    and p_vorige[1] == p_sub_volgende[1])
                                and not (p_volgende[0] == p_sub[0]
                                    and p_volgende[1] == p_sub[1])
                                and not (p_volgende[0] == p_sub_volgende[0]
                                    and p_volgende[1] == p_sub_volgende[1])):
                                if cls.lijn_raakt_lijn(
                                        p_vorige, p_volgende, p_sub,
                                        p_sub_volgende):
                                    over_lijn = True

                        hoek_vorige = cls.bereken_hoek(
                                p_volgende, p_vorige, p,
                                not coordinaten_met_klok_mee)
                        if hoek_vorige > 180:
                            hoek_vorige = 360 - hoek_vorige

                        hoek_volgende = cls.bereken_hoek(
                                p, p_volgende, p_vorige,
                                not coordinaten_met_klok_mee)
                        if hoek_volgende > 180:
                            hoek_volgende = 360 - hoek_volgende

                        if ((not over_lijn) and p[3] < 180
                                and hoek_vorige < p_vorige[3]
                                and hoek_volgende < p_volgende[3]):
                            poly1 = [
                                [p[0], p[1], p[2], p[3]],
                                [p_vorige[0], p_vorige[1],
                                     p_vorige[2], hoek_vorige],
                                [p_volgende[0], p_volgende[1],
                                     p_volgende[2], hoek_volgende]]
                            poly2 = []

                            for i_temp in range(l):
                                p_temp = poly[i_temp]
                                if (not (p[0] == p_temp[0]
                                         and p[1]==p_temp[1])):
                                    if (p_vorige[0] == p_temp[0]
                                             and p_vorige[1] == p_temp[1]):
                                        poly2.append([
                                            p_temp[0],
                                            p_temp[1],
                                            p_temp[2],
                                            (p_temp[3] - hoek_vorige)])
                                    elif (p_volgende[0] == p_temp[0]
                                              and p_volgende[1] == p_temp[1]):
                                        poly2.append([
                                            p_temp[0],
                                            p_temp[1],
                                            p_temp[2],
                                            (p_temp[3] - hoek_volgende)])
                                    else:
                                        poly2.append(p_temp)

                            polys.append(poly1[:])
                            polys.append(poly2[:])
                            driehoeken_lijnen_intern.append([
                                    p_vorige[2], p_volgende[2]])
                            del polys[poly_i]
                            gevonden = True
                            break

                elif l < 3:
                    raise ValueError('opgegeven poly heeft minder dan drie punten')
                else: # l = 3
                    driehoeken.append(poly[:])
                    del polys[poly_i]
                    gevonden = True
                    break

            if gevonden:
                driehoek_iterator(polys)

        coordinaten = [[[x, y, i, h] for i, (x, y, h)
                           in enumerate(np_array.tolist())]]

        driehoek_iterator(coordinaten)

        driehoeken = [[d[0][2], d[1][2], d[2][2]] for d in driehoeken]

        driehoeken, driehoeken_lijnen_intern = \
            cls.optimaliseer_net(np_array, driehoeken, driehoeken_lijnen_intern)

        return np.array(driehoeken), np.array(driehoeken_lijnen_intern)


    @classmethod
    def optimaliseer_net(cls, np_array, driehoeken, driehoeken_lijnen_intern):
        """Triangulatie van niet-convexe polygoon optimatliseren.

        Deze functie wordt gebruikt aan het einde van 'genereer_net' functie.

        Scherpe hoeken kunnen met floating points tot onnauwkeurigheden leiden.
        Als lijn tussen twee driehoeken (vierhoek) een scherpe hoek oplevert,
        dan lijn flippen en andere hoekpunten in deze vierhoek laten verbinden.
        """

        aantal_iteraties = 0

        def iterator_optimaliseren():
            nonlocal driehoeken
            nonlocal driehoeken_lijnen_intern
            nonlocal aantal_iteraties

            aantal_iteraties += 1
            if aantal_iteraties > 9999:
                raise RecursionError('maximale aantal iteraties bereikt')

            gevonden = False
            for i_intern, lijn_intern in enumerate(driehoeken_lijnen_intern):
                # zoek corresponderende driehoeken
                driehoek1 = None
                driehoek2 = None
                for i_driehoek, driehoek in enumerate(driehoeken):
                    if (lijn_intern[0] in driehoek
                            and lijn_intern[1] in driehoek):
                        if driehoek1 is None:
                            driehoek1 = (int(str(i_driehoek)), driehoek[:])
                        else:
                            driehoek2 = (int(str(i_driehoek)), driehoek[:])
                            break
                if driehoek1 is None or driehoek2 is None:
                    continue
                tmp1 = driehoek1[1][:]
                del tmp1[tmp1.index(lijn_intern[0])]
                del tmp1[tmp1.index(lijn_intern[1])]
                tmp2 = driehoek2[1][:]
                del tmp2[tmp2.index(lijn_intern[0])]
                del tmp2[tmp2.index(lijn_intern[1])]

                # zoek waarden van hoekpunten van vierhoek op
                ip1 = lijn_intern[0]
                ip2 = lijn_intern[1]
                ip3 = tmp1[0]
                ip4 = tmp2[0]
                p1 = np_array[ip1][:2].tolist()
                p2 = np_array[ip2][:2].tolist()
                p3 = np_array[ip3][:2].tolist()
                p4 = np_array[ip4][:2].tolist()

                # check of nieuwe diagonaal wel de oude diagonaal kruist
                #    (anders komt nieuwe diagonaal wellicht buiten Vorm
                if not cls.lijn_raakt_lijn(p1, p2, p3, p4):
                    continue

                # bereken hoeken van diagonalen en check of nieuwe groter zijn
                hoek_bestaand_1 = min(cls.bereken_hoek(p1, p2, p3, True),
                                      cls.bereken_hoek(p1, p2, p3, False))
                hoek_bestaand_2 = min(cls.bereken_hoek(p2, p1, p3, True),
                                      cls.bereken_hoek(p2, p1, p3, False))
                hoek_bestaand_3 = min(cls.bereken_hoek(p1, p2, p4, True),
                                      cls.bereken_hoek(p1, p2, p4, False))
                hoek_bestaand_4 = min(cls.bereken_hoek(p2, p1, p4, True),
                                      cls.bereken_hoek(p2, p1, p4, False))
                hoek_nieuw_1 = min(cls.bereken_hoek(p3, p4, p1, True),
                                         cls.bereken_hoek(p3, p4, p1, False))
                hoek_nieuw_2 = min(cls.bereken_hoek(p4, p3, p1, True),
                                         cls.bereken_hoek(p4, p3, p1, False))
                hoek_nieuw_3 = min(cls.bereken_hoek(p3, p4, p2, True),
                                         cls.bereken_hoek(p3, p4, p2, False))
                hoek_nieuw_4 = min(cls.bereken_hoek(p4, p3, p2, True),
                                         cls.bereken_hoek(p4, p3, p2, False))
                min_hoek_bestaand = min(hoek_bestaand_1, hoek_bestaand_2,
                                        hoek_bestaand_3, hoek_bestaand_4)
                min_hoek_nieuw = min(hoek_nieuw_1, hoek_nieuw_2,
                                     hoek_nieuw_3, hoek_nieuw_4)

                if min_hoek_nieuw > min_hoek_bestaand:
                    gevonden = True
                    # vervang huidige driehoeken
                    tmp_driehoeken = []
                    for i, driehoek in enumerate(driehoeken):
                        if i == driehoek1[0]:
                            tmp_driehoeken.append([ip3, ip4, ip1])
                        elif i == driehoek2[0]:
                            tmp_driehoeken.append([ip3, ip4, ip2])
                        else:
                            tmp_driehoeken.append(driehoek)
                    driehoeken = tmp_driehoeken
                    # vervang huidige interne lijn
                    tmp_lijnen_intern = []
                    for i, lijn in enumerate(driehoeken_lijnen_intern):
                        if i == i_intern:
                            tmp_lijnen_intern.append([ip3, ip4])
                        else:
                            tmp_lijnen_intern.append(lijn)
                    driehoeken_lijnen_intern = tmp_lijnen_intern
                    break

            if gevonden:
                iterator_optimaliseren()

        # start proces
        iterator_optimaliseren()

        return driehoeken, driehoeken_lijnen_intern

    
###############################################################################
    
    
class Vorm(pc.BasisObject):
    """Betreft een meetkundig 2D vorm met bijbehorende eigenschappen.

    AANMAKEN VORM
        v1 = Vorm(Lijn)                 invoeren van één Lijn object
        v2 = Vorm([(0,0),(1,1),(1,0)])  direct invoeren knoopcoordinaten

    EENHEID
        v.eenheid       opvragen huidige eenheid; of None als alleen getal
        v.eenheid = 'm' alle waarden in alle knoopobjecten naar 'm'
        v.gebruik_eenheid('m')   zelfde als bovenstaande, retourneert object

    EIGENSCHAPPEN       naam + '_'  -->  Waarde object i.p.v. getal
        v.O             omtrek   (bijv. v.O_ geeft omtrek Waarde met eenheid)
        v.A             oppervlakte
        v.xmin v.xmax   minimum en maximum x-waarde
        v.ymin v.ymax   minimum en maximum y-waarde
        v.ncx  v.ncy    x- en y-waarde normaalkrachtencetrum (zwaartepunt)
        v.Ixx  v.Iyy    oppervlakte traagheidsmoment in x- en y-richting
        v.Ixy           traagheidsproduct (is 0 voor symmetrische vormen)
        v.I1   v.I2     hoofdtraagheidsmomenten (1 sterke richting, 2 zwakke)
        v.alpha         hoek (tegen klok in) hoofdtraagheidsassen
        v.Wxmin v.Wxmax weerstandsmoment voor vezel x-minimaal en x-maximaal
        v.Wymin v.Wymax weerstandsmoment voor vezel y-minimaal en y-maximaal
        v.kxmin v.kxmax laagste/hoogste x-waarde van kern
        v.kymin v.kymax laagste/hoogste y-waarde van kern

    KNOOP COORDINATEN
        v.array                 Numpy array met x/y coordinaten
        v.array_gesloten        zelfde, met kopie 1e knoop aan het einde
        v.kern_array            Numpy array met x/y coordinaten van kern
        v.kern_array_gesloten   zelfde, met kopie 1e knoop aan het einde

    LIJN OBJECT
        v.lijn        genereert een Lijn object van vorm omtrek (gesloten)

    BEWERKINGEN
        v[3]          subset Numpy array object met getallen (zonder eenheid)
        len(v)        aantal knopen
        for k in v:   itereert over knopen, geeft Knoop object (met eenheid)

    OVERIG
        v.plot()                Matplotlib plot met vormeigenschappen
        v.print_eigenschappen() print overzicht van eigenschappen
        v.print_eigenschappen(knopen=True)  zelfde, met lijst van knopen
    """

    fn = VormFuncties()

    EIGENSCHAPPEN = ('O A xmin xmax ymin ymax ncx ncy Ixx Iyy Ixy I1 I2 '
                     'alpha Wxmin Wxmax Wymin Wymax kxmin kxmax kymin kymax '
                     ).split()

    AFRONDEN_NAAR_NUL = 1e-13

    def __init__(self, lijn:Union[pc.Lijn, list, tuple]):
        super().__init__()

        if not isinstance(lijn, pc.Lijn):
            lijn = pc.Lijn(*lijn)

        self._eenheid = lijn.eenheid

        # controleer knoop data en maak Numpy array
        self._array = self._check_knopen(lijn.array.copy())

        # aanmaken vorm eigenschappen
        # normaal een float object, met underscore erachter een Waarde object
        self.O = None       # omtrek
        self.O_ = None
        self.A = None       # oppervlakte
        self.A_ = None
        self.xmin = None    # laagste x-waarde (links)
        self.xmin_ = None
        self.xmax = None    # hoogste x-waarde (rechts)
        self.xmax_ = None
        self.ymin = None    # laagste y-waarde (onder)
        self.ymin_ = None
        self.ymax = None    # hoogste y-waarde (boven)
        self.ymax_ = None
        self.ncx = None     # normaalkrachtcentrum x (zwaartepunt horizontaal)
        self.ncx_ = None
        self.ncy = None     # normaalkrachtcentrum y (zwaartepunt verticaal)
        self.ncy_ = None
        self.Ixx = None     # traagheidsmoment xx (buiging belasting x-richting)
        self.Ixx_ = None
        self.Iyy = None     # traagheidsmoment yy (buiging belasting y-richting)
        self.Iyy_ = None
        self.Ixy = None     # wringtraagheidsmoment xy
        self.Ixy_ = None
        self.I1 = None      # hoofdtraagheidsmoment 1 (sterke as)
        self.I1_ = None
        self.I2 = None      # hoofdtraagheidsmoment 2 (zwakke as)
        self.I1_ = None
        self.alpha = None   # hoek van hoofdtraagheidsassen in graden
        self.alpha_ = None
        self.Wxmin = None   # weerstandsmoment t.p.v. laagste x-waarde (links)
        self.Wxmin_ = None
        self.Wxmax = None   # weerstandsmoment t.p.v. hoogste x-waarde (rechts)
        self.Wxmax_ = None
        self.Wymin = None   # weerstandsmoment t.p.v. laagste y-waarde (onder)
        self.Wymin_ = None
        self.Wymax = None   # weerstandsmoment t.p.v. hoogste y-waarde (boven)
        self.Wymax_ = None
        self.kxmin = None   # laagste x-waarde kern (links)
        self.kxmin_ = None
        self.kxmax = None   # hoogste x-waarde kern (rechts)
        self.kxmax_ = None
        self.kymin = None   # laagste y-waarde kern (onder)
        self.kymin_ = None
        self.kymax = None   # hoogste y-waarde kern (boven)
        self.kymax_ = None

        self._bereken_eigenschappen()
        self._bereken_waardes()

        self._kern_array = np.array([])  # xy coordinaten van kern
        self._bereken_kern_array()

    def _check_knopen(self, np_array):
        """Checkt knopen en berekent hoeken."""

        coordinaten = []
        coordinaten_links_rechts = []

        # zorgen dat laatste knoop niet zelfde is als eerste
        if np_array[0].tolist() == np_array[-1].tolist():
            np_array = np.delete(np_array, (-1), axis=0)

        # zorgen dat omtrek lijnen elkaar nergens kruisen/raken
        l = len(np_array)
        for i in range(l):
            # kies een lijn
            p1 = [np_array[i][0], np_array[i][1]]
            p2 = [np_array[(i+1)%l][0], np_array[(i+1)%l][1]]
            for ii in range(i, l):
                # check alle andere lijnen
                if i != (ii-1)%l and i != ii and i != (ii+1)%l:
                    pp1 = [np_array[ii][0], np_array[ii][1]]
                    pp2 = [np_array[(ii+1)%l][0], np_array[(ii+1)%l][1]]
                    if self.fn.lijn_raakt_lijn(p1, p2, pp1, pp2):
                        raise ValueError('Omtrek van vorm mag zichzelf nergens raken/kruisen. Volgende lijnen doen dat wel: Lijn({}, {}) & Lijn({}, {})'.format(p1, p2, pp1, pp2))

        def verwijder_punt_op_lijn(c):
            nonlocal coordinaten_links_rechts
            punt_op_lijn_gevonden = False
            c = c if (isinstance(c, list) or isinstance(c, tuple)) else []
            n = len(c)

            for i in range(n):
                x = c[i][0]
                y = c[i][1]
                vorige_x = c[(i - 1 + n) % n][0]
                vorige_y = c[(i - 1 + n) % n][1]
                volgende_x = c[(i + 1) % n][0]
                volgende_y = c[(i + 1) % n][1]
                if self.fn.punt_op_lijn([vorige_x, vorige_y],
                                        [volgende_x, volgende_y], [x, y]):
                    del c[i]
                    punt_op_lijn_gevonden = True
                    break
            if punt_op_lijn_gevonden:
                verwijder_punt_op_lijn(c)
            else:
                n = len(c)
                for i in range(n):
                    coordinaten.append([c[i][0], c[i][1], i])
                coordinaten_links_rechts = sorted(
                        coordinaten, key=lambda x: x[0])

        verwijder_punt_op_lijn(np_array.tolist())

        if len(np_array) < 3:
            raise ValueError('Vorm moet minimaal drie knopen bevatten (die niet op één lijn liggen).')

        return np_array

    def _float(self, waarde):
        """Zet object om naar een float en rond hele kleine waarden af naar nul."""
        f = float(waarde)
        if f > -1*self.AFRONDEN_NAAR_NUL and f < 1*self.AFRONDEN_NAAR_NUL:
            f = 0.0
        return f

    def _bereken_eigenschappen(self, alleen_A_O_minmax_nc=False):
        """Berekent alle geometrische eigenschappen."""

        arr = self.array
        Xi = arr[:,0]
        Yi = arr[:,1]
        Xii = np.delete(np.hstack((Xi, np.array([Xi[0]]))), 0)
        Yii = np.delete(np.hstack((Yi, np.array([Yi[0]]))), 0)
        determinant = Xi*Yii - Xii*Yi

        A = 1/2 * sum(determinant)
        if not A:
            return

        # als coordinaten tegen de klok in zijn ingevoerd, dan positief
        # als coordinaten met de klok mee zijn ingevoerd, dan negatief
        teken = 1.0 if A > 0 else -1.0

        self.A = self._float(teken * A)
        self.O = self._float(pc.Lijn(self.array_gesloten.tolist()))
        self.xmin = self._float(Xi.min())
        self.xmax = self._float(Xi.max())
        self.ymin = self._float(Yi.min())
        self.ymax = self._float(Yi.max())
        self.ncx = self._float(
            teken * 1/6/self.A * sum((Xi + Xii) * determinant))
        self.ncy = self._float(
            teken * 1/6/self.A * sum((Yi + Yii) * determinant))

        if alleen_A_O_minmax_nc:
            return

        # vanaf nu alle coordinaten t.o.v. zwaartepunt relativeren
        Xi_ = Xi - self.ncx
        Yi_ = Yi - self.ncy
        Xii_ = np.delete(np.hstack((Xi_, np.array([Xi_[0]]))), 0)
        Yii_ = np.delete(np.hstack((Yi_, np.array([Yi_[0]]))), 0)
        determinant_ = Xi_*Yii_ - Xii_*Yi_

        self.Ixx = self._float(
            teken * 1/12 * sum((Xi_**2 + Xi_*Xii_ + Xii_**2) * determinant_))
        self.Iyy = self._float(
            teken * 1/12 * sum((Yi_**2 + Yi_*Yii_ + Yii_**2) * determinant_))
        self.Ixy = self._float(
            teken * 1/24 * sum((Xi_*Yii_ + 2*Xi_*Yi_ + 2*Xii_*Yii_ + Xii_*Yi_) * determinant_))

        self.Wxmin = self._float(self.Ixx / abs(self.ncx - self.xmin))
        self.Wxmax = self._float(self.Ixx / abs(self.ncx - self.xmax))
        self.Wymin = self._float(self.Iyy / abs(self.ncx - self.xmin))
        self.Wymax = self._float(self.Iyy / abs(self.ncx - self.xmax))
        self.kxmin = self._float(-1 * self.Wxmax / self.A)
        self.kxmax = self._float(self.Wxmin / self.A)
        self.kymin = self._float(-1 * self.Wxmax / self.A)
        self.kymax = self._float(self.Wxmin / self.A)
        self.I1 = self._float((self.Ixx + self.Iyy) / 2
                   + np.sqrt((self.Ixx - self.Iyy)**2 + 4*self.Ixy**2)/2)
        self.I2 = self._float((self.Ixx + self.Iyy) / 2
                   - np.sqrt((self.Ixx - self.Iyy)**2 + 4*self.Ixy**2)/2)
        if self.Ixx - self.Iyy != 0:  # kan niet delen door 0
            self.alpha = self._float((math.atan(2 * self.Ixy /
                (self.Ixx - self.Iyy))/2) * (360 / (2 * math.pi)))
        elif self.Ixy == 0: # hor/vert symmetrisch: bijvoorbeeld rechthoek
            self.alpha = 0.0
        else: # Ixx == Iyy, maar niet hor/vert symmetrisch: bijvoorbeeld ruit
            self.alpha = 45.0

    def _bereken_waardes(self):
        # maak Waarde objecten met eenheid
        oppervlakte_eenheid = None
        weerstand_eenheid = None
        traagheid_eenheid = None
        if self.eenheid is not None:
            oppervlakte_eenheid = '{}2'.format(self.eenheid)
            weerstand_eenheid = '{}3'.format(self.eenheid)
            traagheid_eenheid = '{}4'.format(self.eenheid)
        self.O_ = pc.Waarde(self.O, self.eenheid)
        self.A_ = pc.Waarde(self.A, oppervlakte_eenheid)
        self.xmin_ = pc.Waarde(self.xmin, self.eenheid)
        self.xmax_ = pc.Waarde(self.xmax, self.eenheid)
        self.ymin_ = pc.Waarde(self.ymin, self.eenheid)
        self.ymax_ = pc.Waarde(self.ymax, self.eenheid)
        self.ncx_ = pc.Waarde(self.ncx, self.eenheid)
        self.ncy_ = pc.Waarde(self.ncy, self.eenheid)
        self.Ixx_ = pc.Waarde(self.Ixx, traagheid_eenheid)
        self.Iyy_ = pc.Waarde(self.Iyy, traagheid_eenheid)
        self.Ixy_ = pc.Waarde(self.Ixy, traagheid_eenheid)
        self.I1_ = pc.Waarde(self.I1, traagheid_eenheid)
        self.I2_ = pc.Waarde(self.I2, traagheid_eenheid)
        self.alpha_ = pc.Waarde(self.alpha, 'deg')
        self.Wxmin_ = pc.Waarde(self.Wxmin, weerstand_eenheid)
        self.Wxmax_ = pc.Waarde(self.Wxmax, weerstand_eenheid)
        self.Wymin_ = pc.Waarde(self.Wymin, weerstand_eenheid)
        self.Wymax_ = pc.Waarde(self.Wymax, weerstand_eenheid)
        self.kxmin_ = pc.Waarde(self.kxmin, self.eenheid)
        self.kxmax_ = pc.Waarde(self.kxmax, self.eenheid)
        self.kymin_ = pc.Waarde(self.kymin, self.eenheid)
        self.kymax_ = pc.Waarde(self.kymax, self.eenheid)

    def _bereken_kern_array(self):
        """Berekent het gebied daar waar een normaaldrukkracht NIET in trekspanningen resulteert."""
        kernpunten = [] # lijst met xy coordinaten van kern
        elastiek_lijn = [] # lijst met xy coordinaten als elastiek om vorm

        def bereken_geroteerde_coordinaten(p, alpha, z):
            alpha = -alpha
            xrot = (1.0 * math.cos(alpha * (2 * math.pi / 360)) *
                    (p[0] - z[0]) + math.sin(alpha * (2 * math.pi / 360)) *
                    (p[1] - z[1]))
            yrot = (-1.0 * math.sin(alpha * (2 * math.pi / 360)) *
                    (p[0] - z[0]) + math.cos(alpha * (2 * math.pi / 360)) *
                    (p[1] - z[1]))
            return [xrot + z[0], yrot + z[1]]

        def maak_groot_getal(x):
            return math.floor(x * 1e12)

        c = self.array.tolist()
        l = len(c)
        p = [c[0][0], c[0][1]]

        for i in range(l):
            geldig = True
            p_volgende = [c[(i + 1) % l][0], c[(i + 1) % l][1]]
            test_lijn = [p.copy(), p_volgende.copy()]
            if test_lijn[0][0] == test_lijn[1][0]: # verticale lijn
                if (test_lijn[0][0] < self.xmax
                        and test_lijn[0][0] > self.xmin):
                    geldig = False
            else: # horizontaal of diagonaal
                a = (1.0 * (test_lijn[1][1] - test_lijn[0][1]) /
                    (test_lijn[1][0] - test_lijn[0][0]))
                b = 1.0 * test_lijn[0][1] - test_lijn[0][0] * a
                is_hoger = False
                is_lager = False
                for ii in range(l):
                    if (maak_groot_getal(c[ii][1])
                            > maak_groot_getal(a * c[ii][0] + b) + 10):
                        is_hoger = True
                    elif (maak_groot_getal(c[ii][1])
                            < maak_groot_getal(a * c[ii][0] + b) - 10):
                        is_lager = True
                if is_hoger and is_lager:
                    geldig = False
            if geldig:
                elastiek_lijn.append([p.copy(), p_volgende.copy()])
                p = [c[(i + 1) % l][0], c[(i + 1) % l][1]]

        ncx = self.ncx
        ncy = self.ncy
        n = len(elastiek_lijn)
        for i in range(n):
            rand_lijn = elastiek_lijn[i]
            lijn = [[rand_lijn[0][0] - ncx, rand_lijn[0][1] - ncy],
                    [rand_lijn[1][0] - ncx, rand_lijn[1][1] - ncy]]
            inverse_x1 = 0
            inverse_y1 = 0
            if lijn[0][0] == lijn[1][0]: # verticale lijn
                if lijn[0][0] == 0:
                    inverse_x1 = 99e99
                else:
                    inverse_x1 = 1 / lijn[0][0]
                inverse_y1 = 0
            elif lijn[0][1] == lijn[1][1]: # horizontale lijn
                inverse_x1 = 0
                if lijn[0][1] == 0:
                    inverse_y1 = 99e99
                else:
                    inverse_y1 = 1/ lijn[0][1]
            else: # diagonale lijn
                lijn = sorted(lijn, key=lambda x: x[0]) # sorteer op x-waarde
                a = (1.0 * (lijn[1][1] - lijn[0][1])
                     / (lijn[1][0] - lijn[0][0]))
                b = 1.0 * lijn[0][1] - lijn[0][0] * a
                if a == 0 or b == 0:
                    inverse_x1 = 99e99
                else:
                    inverse_x1 = 1/ (-1 * b / a)
                if b == 0:
                    inverse_y1 = 99e99
                else:
                    inverse_y1 = 1 / b
            ex = (-1.0 / self.A * (self.Ixx * inverse_x1
                                   + self.Ixy * inverse_y1))
            ey = (-1.0 / self.A * (self.Ixy * inverse_x1
                                   + self.Iyy * inverse_y1))
            kernpunten.append([ex + ncx, ey + ncy])

        self._kern_array = np.array(kernpunten)

    @property
    def eenheid(self) -> str:
        """Geeft eenheid van Waarde. 'None' als geen eenheid."""
        return self._eenheid

    @eenheid.setter
    def eenheid(self, eenheid:str):
        """Zet knopen om naar nieuwe eenheid."""
        tmp_knopen = []
        oude_eenheid = self._eenheid

        for k_array in self.array:
            k = pc.Knoop(k_array.tolist())
            k.eenheid = oude_eenheid
            k.eenheid = eenheid
            tmp_knopen.append(k.array.tolist())

        self._array = self._check_knopen(np.array(tmp_knopen, dtype='float64'))
        self._eenheid = eenheid

        self._bereken_eigenschappen()
        self._bereken_waardes()
        self._bereken_kern_array()

    def gebruik_eenheid(self, eenheid:str):
        """Zet knopen om naar nieuwe eenheid en retourneert object."""
        self.eenheid = eenheid
        return self

    @property
    def array(self) -> np.array:
        """Retourneert Numpy array object met alle getallen (zonder eenheid)."""
        return self._array

    @property
    def array_gesloten(self) -> np.array:
        """Retourneert Numpy array object met alle getallen (zonder eenheid) waarbij startpunt OOK als laatste punt wordt aangehouden."""
        if len(self.array) > 0:
            return np.append(self.array, [self[0].tolist()], axis=0)
        else:
            return np.array([])

    @property
    def kern_array(self) -> np.array:
        """Retourneert Numpy array object met alle knopen van kern (zonder eenheid)."""
        return self._kern_array

    @property
    def kern_array_gesloten(self) -> np.array:
        """Retourneert Numpy array object met allen knopen van kern (zonder eenheid) waarbij startpunt OOK als laatste punt wordt aangehouden."""
        if len(self.kern_array) > 0:
            return np.append(self.kern_array,
                         [self.kern_array[0].tolist()], axis=0)
        else:
            return np.array([])

    @property
    def lijn(self) -> pc.Lijn:
        """Retourneert Lijn object dat correspondeert met (gesloten) Vorm."""
        knoop_objecten = [k for k in self]
        return pc.Lijn(knoop_objecten + [knoop_objecten[0]])

    def plot(self):
        """Teken vorm."""
        plt.axis('equal')

        # omtrek vorm
        if len(self.array > 2):
            X = self.array_gesloten[:,0]
            Y = self.array_gesloten[:,1]
            plt.fill(X, Y, 'r', alpha=0.2)
            plt.plot(X, Y, 'r-', lw=2)

        # kern vorm
        if len(self.kern_array > 2):
            X = self.kern_array_gesloten[:,0]
            Y = self.kern_array_gesloten[:,1]
            plt.fill(X, Y, 'b', alpha=0.2)
            plt.plot(X, Y, 'b-', lw=1)

        # hoofdtraagsheidsassens
        marge = 0.1
        hoek = self.alpha / 180 * math.pi
        ymin = self.ymin - (self.ymax - self.ymin) * marge
        ymax = self.ymax + (self.ymax - self.ymin) * marge
        xmin = self.xmin - (self.xmax - self.xmin) * marge
        xmax = self.xmax + (self.xmax - self.xmin) * marge
        x_ymin = self.ncx + (self.ncy - ymin) * math.tan(hoek)
        x_ymax = self.ncx - (ymax - self.ncy) * math.tan(hoek)
        y_xmin = self.ncy - (self.ncx - xmin) * math.tan(hoek)
        y_xmax = self.ncy + (xmax - self.ncx) * math.tan(hoek)
        plt.plot([x_ymin, x_ymax], [ymin, ymax], 'g-', lw=1)
        plt.plot([xmin, xmax], [y_xmin, y_xmax], 'g-', lw=1)

        # zwaartepunt
        plt.plot(self.ncx, self.ncy, 'bo')

        plt.show()

    def print_eigenschappen(self, knopen=False):
        print_queue = []

        if knopen:
            print_queue.append(
                    'knopen (afgerond op 2 decimalen):\n{} {}'.format(
                    [(round(k[0], 2), round(k[1], 2)) for k in self.array],
                    self.eenheid if self.eenheid is not None else '').strip())

            print_queue.append('')

        print_queue.append('\n'.join(['{:>8} = {:.3f}'.format(
                a, getattr(self, a+'_')) for a in self.EIGENSCHAPPEN]))

        print('\n{}\n'.format('\n'.join(print_queue)))

    def __getitem__(self, index) -> np.array:
        """Retourneert subset Numpy array object met getallen (zonder eenheid)."""
        return self.array[index]

    def __len__(self):
        """Geeft aantal knopen."""
        return len(self.array)

    def __iter__(self):
        """Itereert over knopen en geeft Knoop objecten terug."""
        for k_array in self.array:
            k = pc.Knoop(k_array.tolist())
            k.eenheid = self.eenheid
            yield k

    def __repr__(self):
        cls_naam = type(self).__name__
        knopen = ', '.join(repr(k) for k in self)
        return '{}(Lijn({}))'.format(cls_naam, knopen)

    def __str__(self):
        return '{} {}'.format(
                # coordinaten afgerond op twee decimalen
                [(round(k[0], 2), round(k[1], 2)) for k in self.array],
                self.eenheid if self.eenheid is not None else '').strip()


if __name__ == '__main__':

    print('v1 - lijn met cirkelboog en bezier')
    v1 = Vorm(pc.Lijn(
            (4, -5), (-10, 10)
        ).lijn_cirkelboog(
            middelpunt=(0,0),
            gradenhoek=+220
        ).lijn_recht(
            naar=(4, 10)
        ).lijn_bezier(
            richting=(-10,-4),
            naar=(4, -5)
        ).transformeren(
            rotatiehoek=30,
            translatie=[15, 5],
        ))
    v1.plot()
    v1.print_eigenschappen()

    print('v2 - polygon')
    v2 = Vorm([[0,0], [0,10], [4,10], [4,7], [6,7], [6,10], [10,10], [10,0]])
    v2.plot()
    v2.print_eigenschappen()

    print('v3 - cirkel benaderd d.m.v. een lijn met cirkelboog')
    v3= Vorm(pc.Lijn([-1,0]).lijn_cirkelboog(middelpunt=(0,0), gradenhoek=360))
    v3.plot()
    v3.print_eigenschappen()

    print('v4 - driehoek')
    v4 = Vorm(pc.Lijn([0,0], [6,-3], [10,4]).gebruik_eenheid('cm'))
    v4.plot()
    v4.print_eigenschappen()

    print('v5 - rechthoek met transformatie')
    v5 = Vorm(pc.Lijn([-50,-20], [50,-20], [50,20], [-50, 20]).transformeren(
              rotatiepunt=None, # bij None: neemt zwaartepunt
              rotatiehoek=45, # graden tegen de klok in
              schaalfactor=[1, 1], # vergroten om rotatiepunt; negatief:spiegelen
              translatie=[0, 0] # verplaatsing
        ).gebruik_eenheid('mm'))
    v5.plot()
    v5.print_eigenschappen()

    print('v6 - gedraaide rechthoek met eenheid')
    v6 = Vorm(pc.Lijn([2,0], [0,2], [8,10], [10, 8]).gebruik_eenheid('cm'))
    v6.plot()
    v6.print_eigenschappen()

#     print('v7 - Rechthoek object')
#     v7 = pc.Rechthoek(breedte=30, hoogte=50)
#     v7.plot()
#     v7.print_eigenschappen()

#     print('v8 - Cirkel object')
#     v8 = pc.Cirkel(straal=pc.Waarde(1).dm).gebruik_eenheid('m')
#     v8.plot()
#     v8.print_eigenschappen()



from typing import Union

import numpy as np
import matplotlib.pyplot as plt

import pyco.basis
import pyco.waarde
import pyco.knoop
import pyco.lijst

class pc:
    BasisObject = pyco.basis.BasisObject
    Waarde = pyco.waarde.Waarde
    Lijst = pyco.lijst.Lijst
    Knoop = pyco.knoop.Knoop


class Lijn(pc.BasisObject):
    """Bevat een collectie met knopen, waartussen zich rechte lijnen bevinden.

    AANMAKEN LIJN               invoeren van één of meedere Knoop objecten
        Lijn(Knoop(Waarde(1).cm, Waarde(2).cm)))    begin Knoop object
        Lijn([1,2]) of Lijn((1,2))                  alleen begincoordinaat
        Lijn((1,2), (3,4), (5,6))                   alle knoopcoordinaten

    AANPASSEN EENHEID
        l = Lijn((1,2), (3,4))
        l.eenheid               opvragen huidige eenheid; in dit geval None
        l.eenheid = 'm'         alle waarden in alle knoopobjecten naar 'm'
        l.gebruik_eenheid('m')  zelfde als bovenstaande, retourneert object

    OMZETTEN LIJN NAAR TEKST    resulteert in nieuw string object
        tekst = str(l)          of automatisch met bijvoorbeeld print(l)
        tekst = format(l,'.2f') format configuratie meegeven voor getal

    VERLENGEN LIJN              vanuit laatste knoop (of enige beginknoop)
        l.lijn_recht(naar=(3,4))
            rechte lijn naar een nieuwe knoop

        l.lijn_bezier(richting=(3,4), naar=(5,6), stappen=100)
            (kwadratische) Bezier kromme (met één richtingspunt) naar nieuwe
            knoop waarbij de kromme lijn omgezet wordt in aantal (stappen)
            rechte lijnen; standaard 100 stappen

        l.lijn_cirkelboog(middelpunt=(3,4), gradenhoek=-90, stappen=100)
            cirkelboog met opgegeven cirkel middelpunt over aantal opgegeven
            graden (waarbij 360 is gehele cirkel tekenen; positief is tegen
            klok in; negatief getal is met de klok mee) waarbij kromme lijn
            omgezet wordt in aantal rechte lijnen; standaard 100 stappen

    TRANSFORMEREN LIJN
        l.transformeren(       # standaard zijn alle parameters None
            rotatiepunt=[0,0], # xy Knoop/list; als None dan zwaartepunt vorm
            rotatiehoek=20,    # in graden, positief is tegen de klok in
            schaalfactor=2,  # Waarde/getal: vergrootfactor t.o.v. rotatiepunt
            schaalfactor=[2,3],# of Lijst/list met x-schaalfactor en y-factor
            schaalfactor=[1,-1],# verticaal spiegelen
            schaalfactor=[-1,1],# horizontaal spiegelen
            schaalfactor=[-5,3],# bovenstaande combineren
            translatie=[10,5], # xy verschuiven (na roteren en schalen)
        )

    MOGELIJKE BEWERKINGEN
        waarde = abs(l)         berekent lengte lijnstukken -> Waarde object
        getal = float(l)        berekent lengte lijnstukken -> float object
        for w in v1:            itereert en geeft Knoop object terug
        getal = len(v1)         geeft aantal knopen terug

    NUMPY BEWERKINGEN               gebruikt array object
        2D numpy_array = l1.array   retourneert volledige Numpy array object
                                      (bevat allen getallen, zonder eenheid)
        1D_numpy_array = l1[2]      retourneert knoopcoordinaten op index
        2D_numpy_array = l1[1:3]    retourneert knoopcoordinaten vanuit slice

    WAARDEN VERGELIJKEN         resulteert in een boolean (True/False)
        l1 == l2                is gelijk aan
        l1 != l2                is niet gelijk aan
        l1 &  l2                eenheden zijn zelfde type

    EXTRA OPTIES
        l.plot()                plot simpele 2D weergave van lijn
        l.plot3D()              plot simpele 3D weergave van lijn

        Lijn((4, -5), (-10, 10)).lijn_cirkelboog(middelpunt=(0,0),
            gradenhoek=+220, stappen=50).lijn_recht(naar=(4, 10)).lijn_bezier(
            richting=(-10,-4), naar=(4, -5)).plot()
    """

    def __init__(self, *knopen):
        super().__init__()

        tmp_knopen = []
        self._eenheid = None

        if (len(knopen) == 1
                and (
                    isinstance(knopen[0], list) or
                    isinstance(knopen[0], tuple))
                and len(knopen[0]) > 0
                and (
                    isinstance(knopen[0][0], pc.Knoop) or
                    isinstance(knopen[0][0], list) or
                    isinstance(knopen[0][0], tuple))):
            knopen = knopen[0]

        if len(knopen) < 1:
            raise ValueError('voer minimaal één knoop in')

        for i, knoop in enumerate(knopen):
            if not (isinstance(knoop, pc.Knoop)
                    or isinstance(knoop, list) or isinstance(knoop, tuple)):
                raise TypeError('opgegeven argument is geen Knoop object of lijst met getallen/Waardes')
            if not isinstance(knoop, pc.Knoop):
                knoop = pc.Knoop(knoop)
            if i > 0:
                if ((self._eenheid is None and knoop.eenheid is not None)
                        or (self._eenheid is not None and knoop.eenheid is None)):
                    raise ValueError('knopen moeten zelfde type eenheid hebben')
                knoop.eenheid = self._eenheid
            if i == 0 or (i > 0 and tmp_knopen[-1] != knoop.array.tolist()):
                tmp_knopen.append(knoop.array.tolist())
            if i == 0:
                self._eenheid = knoop.eenheid

        self._array = np.array(tmp_knopen, dtype='float64')

    @property
    def eenheid(self) -> str:
        """Geeft eenheid van Waarde. 'None' als geen eenheid."""
        return self._eenheid

    @eenheid.setter
    def eenheid(self, eenheid:str):
        """Zet knopen om naar nieuwe eenheid."""
        tmp_knopen = []
        oude_eenheid = self._eenheid

        for k_array in self.array:
            k = pc.Knoop(k_array.tolist())
            k.eenheid = oude_eenheid
            k.eenheid = eenheid
            tmp_knopen.append(k.array.tolist())

        self._array = np.array(tmp_knopen, dtype='float64')
        self._eenheid = eenheid

    def gebruik_eenheid(self, eenheid:str):
        """Zet knopen om naar nieuwe eenheid en retourneert object."""
        self.eenheid = eenheid
        return self

    @property
    def array(self) -> np.array:
        """Retourneert Numpy array object met alle getallen (zonder eenheid)."""
        return self._array

    def lijn_recht(self, naar:pc.Knoop):
        """Verlengt lijn object met een extra rechte lijn naar opgegeven knoop."""
        if not (isinstance(naar, pc.Knoop)
                or isinstance(naar, list) or isinstance(naar, tuple)):
            raise TypeError('opgegeven argument is geen Knoop object of lijst met getallen/Waardes')

        if not isinstance(naar, pc.Knoop):
            naar = pc.Knoop(naar)

        if ((self.eenheid is None and naar.eenheid is not None)
                or (self.eenheid is not None and naar.eenheid is None)):
            raise ValueError('knopen moeten zelfde type eenheid hebben')
        naar.eenheid = self.eenheid

        if self[-1].tolist() != naar.array.tolist():
            self._array = np.append(self.array, [naar.array.tolist()], axis=0)
        return self

    def lijn_bezier(self, richting:pc.Knoop, naar:pc.Knoop, stappen:int=100):
        """Verlengt lijn object als kwadratische Bezier kromme naar opgegeven knoop. Hierbij worden <aantal stappen> rechte lijnen gemaakt."""
        if not (isinstance(richting, pc.Knoop)
                or isinstance(richting, list) or isinstance(richting, tuple)):
            raise TypeError('opgegeven richting-knoop is geen Knoop object of lijst met getallen/Waardes')
        if not (isinstance(naar, pc.Knoop)
                or isinstance(naar, list) or isinstance(naar, tuple)):
            raise TypeError('opgegeven naar-knoop is geen Knoop object of lijst met getallen/Waardes')

        if not isinstance(richting, pc.Knoop):
            richting = pc.Knoop(richting)
        if not isinstance(naar, pc.Knoop):
            naar = pc.Knoop(naar)

        if ((self.eenheid is None and richting.eenheid is not None)
                or (self.eenheid is not None and richting.eenheid is None)):
            raise ValueError('knopen moeten zelfde type eenheid hebben')
        richting.eenheid = self.eenheid
        if ((self.eenheid is None and naar.eenheid is not None)
                or (self.eenheid is not None and naar.eenheid is None)):
            raise ValueError('knopen moeten zelfde type eenheid hebben')
        naar.eenheid = self.eenheid

        start = pc.Knoop(self[-1].tolist())
        start.eenheid = self.eenheid

        tt = np.linspace(0, 1, num=stappen+1)
        P1 = start[:]
        P2 = richting[:]
        P3 = naar[:]

        extra_knopen=[]
        for t in tt[1:]:
            # kwadratische Bezier kromme:
            extra_knoop = (1 - t**2)*P1 + 2*t*(1-t)*P2 + t**2*P3
            extra_knopen.append(extra_knoop.tolist())

        self._array = np.append(self.array, extra_knopen, axis=0)
        return self

    def lijn_cirkelboog(self, middelpunt:pc.Knoop, gradenhoek:float, stappen:int=100):
        """Verlengt lijn object als cirkel met middelpunt over x aantal graden. Hierbij worden <aantal stappen> rechte lijnen gemaakt. Positief is tegen klok in, negatief is met de klok mee."""
        if not (isinstance(middelpunt, pc.Knoop)
                or isinstance(middelpunt, list) or isinstance(middelpunt, tuple)):
            raise TypeError('opgegeven middelpunt-knoop is geen Knoop object of lijst met getallen/Waardes')

        if not isinstance(middelpunt, pc.Knoop):
            middelpunt = pc.Knoop(middelpunt)

        if ((self.eenheid is None and middelpunt.eenheid is not None)
                or (self.eenheid is not None and middelpunt.eenheid is None)):
            raise ValueError('knopen moeten zelfde type eenheid hebben')
        middelpunt.eenheid = self.eenheid

        start = pc.Knoop(self[-1].tolist())
        start.eenheid = self.eenheid

        P1 = start[:]
        M = middelpunt[:]
        straal = np.linalg.norm(M - P1)
        hoek = gradenhoek%360 if gradenhoek >= 0 else -1*((-1*gradenhoek)%360)
        hoek = 360 if hoek == 0 else hoek
        start_hoek = (270-np.rad2deg(np.arctan2(M[0]-P1[0], M[1]-P1[1])))%360
        eind_hoek = start_hoek + hoek

        tt = np.linspace(np.deg2rad(start_hoek), np.deg2rad(eind_hoek), num=stappen+1)
        extra_knopen=[]
        for t in tt[1:]:
            extra_knoop = [M[0] + straal*np.cos(t), M[1] + straal*np.sin(t)]
            extra_knopen.append(extra_knoop)

        self._array = np.append(self.array, extra_knopen, axis=0)
        return self

    def transformeren(self,
                      rotatiepunt:pc.Knoop=None,
                      rotatiehoek:Union[pc.Waarde, float, int]=None,
                      schaalfactor:Union[pc.Waarde, float, int,
                                         pc.Lijst, list, tuple]=None,
                      translatie:Union[pc.Lijst, list, tuple]=None):
        """Verplaatst, roteert en/of verschaalt de lijn."""

        if rotatiepunt is not None and isinstance(rotatiepunt, pc.Knoop):
            rotatiepunt = rotatiepunt.array.tolist()
        elif (rotatiepunt is not None
                and (isinstance(rotatiepunt, tuple)
                or isinstance(rotatiepunt, list))):
            rotatiepunt = [rotatiepunt[0], rotatiepunt[1]]
        else:
            # als geen geldige coordinaten, dan bereken zwaartepunt
            Xi = self.array[:,0]
            Yi = self.array[:,1]
            Xii = np.delete(np.hstack((Xi, np.array([Xi[0]]))), 0)
            Yii = np.delete(np.hstack((Yi, np.array([Yi[0]]))), 0)
            determinant = Xi*Yii - Xii*Yi
            A = 1/2 * sum(determinant)
            if not A:
                return
            teken = 1.0 if A > 0 else -1.0
            A_ = float(teken * A)
            ncx = float(teken * 1/6/A_ * sum((Xi + Xii) * determinant))
            ncy = float(teken * 1/6/A_ * sum((Yi + Yii) * determinant))
            rotatiepunt = [ncx, ncy]

        if rotatiehoek is not None:
            tmp_array = self.array - rotatiepunt

            rh = np.radians(float(rotatiehoek))
            rotatie_matrix = np.array([[np.cos(rh), -np.sin(rh)],
                                       [np.sin(rh),  np.cos(rh)]])
            tmp_array = np.matmul(rotatie_matrix, tmp_array.T).T

            self._array = tmp_array + rotatiepunt

        if schaalfactor is not None:
            tmp_array = self.array - rotatiepunt

            if ((isinstance(schaalfactor, list)
                    or isinstance(schaalfactor, tuple)
                    or isinstance(schaalfactor, pc.Lijst))
                    and len(schaalfactor) >= 2):
                sf1 = float(schaalfactor[0])
                sf2 = float(schaalfactor[1])
            else:
                sf1 = float(schaalfactor)
                sf2 = float(schaalfactor)

            schaal_matrix = np.array([[sf1,  0],
                                      [ 0, sf2]])
            tmp_array = np.matmul(schaal_matrix, tmp_array.T).T

            self._array = tmp_array + rotatiepunt

        if translatie is not None and len(translatie) >= 2:
            self._array = self.array + np.array(
                    [translatie[0], translatie[1]])

        return self

    def plot(self):
        """Teken simpele plot van lijn (met 2 dimensies)."""
        plt.plot(self.array[:,0], self.array[:,1])
        plt.axis('equal')
        plt.show()
        
    def plot3D(self):
        """Teken simpele plot van lijn (met 3 dimensies)."""
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = Axes3D(fig)
        plt.plot(self.array[:,0], self.array[:,1], self.array[:,2])
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        plt.show()

    def __eq__ (self, andere):
        """Vergelijkt twee Lijn objecten."""
        if len(self) != len(andere):
            return False
        return all(s == a for s, a in zip(self, andere))

    def __neq__ (self, andere):
        """Vergelijkt negatief twee Lijn objecten."""
        return not self == andere

    def __and__(self, andere):
        """Controleert of Lijn zelfde type eenheid heeft als andere."""
        if not isinstance(andere, pc.Lijn):
            raise TypeError('tweede waarde is geen Lijn object')
        return pc.Waarde(1, self.eenheid) & Waarde(1, andere.eenheid)

    def __iter__(self):
        """Itereert over knopen en geeft Knoop objecten terug."""
        for k_array in self.array:
            k = pc.Knoop(k_array.tolist())
            k.eenheid = self.eenheid
            yield k

    def __getitem__(self, index):
        """Retourneert subset Numpy array object met getallen (zonder eenheid)."""
        return self.array[index]

    def __len__(self):
        """Geeft aantal knopen."""
        return len(self.array)

    def __float__(self):
        """Berekent de totale lengte van de lijnstukken als float object."""
        if len(self) < 2:
            return 0.0
        laatste_kn_arr = self[0]
        lengte = 0.0
        for kn_arr in self[1:]:
            x1 = laatste_kn_arr[0]
            x2 = kn_arr[0]
            y1 = laatste_kn_arr[1]
            y2 = kn_arr[1]
            lengte += np.sqrt((x2-x1)**2 + (y2-y1)**2)
            laatste_kn_arr = kn_arr
        return lengte

    def __abs__(self):
        """Berekent de totale lengte van de lijnstukken als Waarde object."""
        return pc.Waarde(float(self), self.eenheid)

    def __format__(self, config:str=None):
        """Geeft tekst met geformatteerd getal en eenheid."""
        if config is None:
            return str(self)
        knopen = ', '.join(format(k, config).rsplit(')', 1)[0]+')' for k in self)
        eenheid = self.eenheid if self.eenheid is not None else ''
        return '({}) {}'.format(knopen, eenheid).strip()

    def __repr__(self):
        cls_naam = type(self).__name__
        knopen = ', '.join(repr(k) for k in self)
        return '{}({})'.format(cls_naam, knopen)

    def __str__(self):
        knopen = ', '.join(str(k).rsplit(')', 1)[0]+')' for k in self)
        eenheid = self.eenheid if self.eenheid is not None else ''
        return '({}) {}'.format(knopen, eenheid).strip()



import matplotlib.pyplot as plt
import matplotlib.path as mpath
import matplotlib.patches as mpatches

from typing import Union
import base64
from io import BytesIO, StringIO

import numpy as np

# verbergen van waarschuwingen van Matplotlib
import warnings
warnings.filterwarnings('ignore')

import pyco.basis
# import pyco.waarde
# import pyco.lijst
import pyco.lijn
import pyco.venster

class pc:
    BasisObject = pyco.basis.BasisObject
    # Waarde = pyco.waarde.Waarde
    # Lijst = pyco.lijst.Lijst
    Lijn = pyco.lijn.Lijn
    FiguurVenster = pyco.venster.FiguurVenster
    BestandsnaamVenster = pyco.venster.BestandsnaamVenster


class Figuur(pc.BasisObject):
    """
    Tekent een figuur met behulp van matplotlib bibliotheek.

    WERKWIJZE                   plaats functies achter elkaar
    f = Figuur(                 # configuratie
            breedte=7,
            hoogte=7,
        ).lijn(                 # onderdeel 1
            coordinaten=((2, 0), (3, 7)),
        ).fx(                   # onderdeel 2  etc.
            functie = lambda x: x**2,
            x = (-2, 3),
        )                       

    LIJN OBJECT             in plaats van tuple/list met coordinaten,
                            mag ook Lijn object invoeren
        Figuur().punt(
            coordinaten=Lijn([1,2],[3,4],[5,6])
        )

    CONFIGURATIE
        Figuur(
            breedte=7,
            hoogte=7,
            raster=True,
            legenda=True,
            titel='Kunst',
            x_as_titel='variabele $x$',
            y_as_titel='resultaat $y$',
            gelijke_assen=True,
            verberg_assen=False,
            x_as_log=False,
            y_as_log=False,
        )

    ONDERDELEN
        .lijn(
            coordinaten=((2, 0), (3, 7), (-2, 4), (2, 0)),
            breedte=3,
            kleur='red',   # None is auto kleur
            vullen=False,
            arcering='/',
            naam='dit is een lijn',
        )

        .punt(
            coordinaten=((2, 0), (3, 7), (-2, 4)),
            breedte=10,
            kleur='gold',   # None is auto kleur
            stijl='>',
            naam='dit zijn punten',
        )

        .tekst(
            coordinaten=((2, 0), (3, 7), (-2, 4)),
            teksten=('punt 1', 'punt 2', 'punt 3'),
            kleur='brown',   # None is auto kleur
            tekst_grootte='large', # {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
            tekst_font='sans-serif', # {FONTNAME, 'serif', 'sans-serif', 'cursive', 'fantasy', 'monospace'}
            tekst_stijl='normal', # {'normal', 'italic', 'oblique'}
            tekst_gewicht='bold', # {'ultralight', 'light', 'normal', 'regular', 'book', 'medium', 'roman', 'semibold', 'demibold', 'demi', 'bold', 'heavy', 'extra bold', 'black'}
            hor_uitlijnen='center', # {'center', 'right', 'left'}
            vert_uitlijnen='center', # {'center', 'top', 'bottom', 'baseline', 'center_baseline'}
            roteren=30, # in graden
        )

        .kolom(
            coordinaten=((3, 8), (6, 4), (8, 1)),
            kleur='pink',   # None is auto kleur
            breedte=0.4,
            lijn_kleur='peru',
            lijn_breedte=2,
            naam='bar plot',
        )

        .fx(
            functie = lambda x: x**2 + 3*pc.sin(x),
            x = (-2, 3),
            breedte = 2,
            kleur = 'green',   # None is auto kleur
            naam = 'sinus parabool',
        )
        
    WEERGAVE FIGUUR
        f.plot()                # plot direct in notebook/console
        f.plot_venster()        # plot in een popup venster

    OVERIG
        f.volgende_kleur        # gebruik deze eigenschap om verschillende automatische kleuren te gebruiken
        f.png_url               # data url encoded
        f.png_html              # HTML IMG code van PNG afbeelding (data url encoded)
        f.svg_html              # SVG code voor inline HTML gebruik
        f.bewaar_als_png()      # vraag bestandsnaam om op te slaan als PNG
        f.bewaar_als_svg()      # vraag bestandsnaam om op te slaan als SVG

    """

    def __init__(self,
                 breedte=8,
                 hoogte=8,
                 raster:bool=False,
                 legenda:bool=False,
                 titel:str='',
                 x_as_titel:str='',
                 y_as_titel:str='',
                 gelijke_assen:bool=False,
                 verberg_assen:bool=False,
                 x_as_log:bool=False,
                 y_as_log:bool=False):

        super().__init__()

        self.fig, self.ax = plt.subplots(1, 1, figsize=(breedte, hoogte))

        self.min_x = None
        self.max_x = None
        self.min_y = None
        self.max_y = None

        self.marge_x = 0.05
        self.marge_y = 0.05

        self.maak_raster=raster
        self.maak_legenda=legenda

        if self.maak_raster:
           self.ax.axhline(y=0, color='darkgrey', linewidth=1.2)
           self.ax.axvline(x=0, color='darkgrey', linewidth=1.2)

        self.titel = titel
        if self.titel != '':
            self.ax.set_title(self.titel)
        if x_as_titel != '':
            self.ax.set_xlabel(x_as_titel)
        if y_as_titel != '':
            self.ax.set_ylabel(y_as_titel)

        if gelijke_assen:
            self.ax.axis('equal')

        if verberg_assen:
            self.ax.set_axis_off()

        if x_as_log:
            self.ax.set_xscale('log')

        if y_as_log:
            self.ax.set_yscale('log')

        # if True:
        #     self.ax.spines['top'].set_visible(False)

        self._kleuren = plt.rcParams["axes.prop_cycle"]()

        self.alleen_lezen = False

    def _check_coordinaten(self, coordinaten:Union[list, tuple, pc.Lijn]):
        """Controleert coordinaten en bepaalt globaal minimum en maximum."""
        if not isinstance(coordinaten, list) and not isinstance(coordinaten, tuple):
            if isinstance(coordinaten, pc.Lijn):
                coordinaten.array.tolist()
            else:
                raise ValueError('coordinaten is geen Lijn, lijst of tupel')
        for coordinaat in coordinaten:
            if len(coordinaat) != 2:
                raise ValueError('coordinaat heeft minder of meer dan 2 elementen: {}'.format(coordinaat))

        x_waarden = [x for x, _ in coordinaten]
        y_waarden = [y for _, y in coordinaten]

        if self.min_x is not None:
            x_waarden = x_waarden + [self.min_x]
        if self.max_x is not None:
            x_waarden = x_waarden + [self.max_x]
        if self.min_y is not None:
            y_waarden = y_waarden + [self.min_y]
        if self.max_x is not None:
            y_waarden = y_waarden + [self.max_y]

        self.min_x = min(x_waarden)
        self.max_x = max(x_waarden)
        self.min_y = min(y_waarden)
        self.max_y = max(y_waarden)

        return coordinaten

    def _check_alleen_lezen(self):
        if self.alleen_lezen:
            raise Exception('figuur is al afgerond en is daardoor alleen-lezen')

    @property
    def volgende_kleur(self):
        return next(self._kleuren)["color"]

    def lijn(self,
             coordinaten:Union[list, tuple, pc.Lijn],
             breedte=1,
             kleur=None,
             vullen=False,
             arcering='',
             naam:str=None):
        """Trekt een lijn door middel van coordinaten."""

        self._check_alleen_lezen()
        
        kleur = self.volgende_kleur if kleur is None else kleur

        coordinaten = self._check_coordinaten(coordinaten)
        codes = [mpath.Path.MOVETO, *[mpath.Path.LINETO for _ in range(len(coordinaten)-1)]]
        path = mpath.Path(coordinaten, codes)
        patch = mpatches.PathPatch(path,
                                   linewidth=breedte,
                                   hatch=arcering,
                                   fill=vullen,
                                   color=kleur,
                                   label=naam)
        self.ax.add_patch(patch)
        return self

    def punt(self,
             coordinaten:Union[list, tuple, pc.Lijn],
             breedte=1,
             kleur='black',
             stijl='o',
             naam:str=None):
        """Plot punten door middel van coordinaten."""

        self._check_alleen_lezen()

        coordinaten = self._check_coordinaten(coordinaten)
        X = [x for x, _ in coordinaten]
        Y = [y for _, y in coordinaten]
        self.ax.scatter(X, Y,
                        color=kleur,
                        marker=stijl,
                        s=breedte*30,
                        label=naam)
        return self

    def tekst(self,
             coordinaten:Union[list, tuple, pc.Lijn],
             teksten:Union[list, tuple],
             kleur=None,
             tekst_grootte:str='medium', # {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
             tekst_font:str='sans-serif', # {FONTNAME, 'serif', 'sans-serif', 'cursive', 'fantasy', 'monospace'}
             tekst_stijl:str='normal', # {'normal', 'italic', 'oblique'}
             tekst_gewicht:str='normal', # {'ultralight', 'light', 'normal', 'regular', 'book', 'medium', 'roman', 'semibold', 'demibold', 'demi', 'bold', 'heavy', 'extra bold', 'black'}
             hor_uitlijnen:str='center', # {'center', 'right', 'left'}
             vert_uitlijnen:str='center', # {'center', 'top', 'bottom', 'baseline', 'center_baseline'}
             roteren:Union[float, int, str]=0): # float in degrees
        """Laat teksten zien op posities gegeven door coordinaten."""

        self._check_alleen_lezen()
        
        kleur = self.volgende_kleur if kleur is None else kleur

        coordinaten = self._check_coordinaten(coordinaten)
        X = [x for x, _ in coordinaten]
        Y = [y for _, y in coordinaten]
        for i, tekst in enumerate(teksten):
            x = X[i]
            y = Y[i]
            self.ax.text(x, y,
                            s=tekst,
                            color=kleur,
                            fontsize=tekst_grootte,
                            fontfamily=tekst_font,
                            fontstyle=tekst_stijl,
                            fontweight=tekst_gewicht,
                            horizontalalignment=hor_uitlijnen,
                            verticalalignment=vert_uitlijnen,
                            rotation=roteren)
        return self

    def kolom(self,
             coordinaten:Union[list, tuple, pc.Lijn],
             kleur:str=None,
             breedte:Union[float, int]=0.8,
             lijn_kleur:str='black',
             lijn_breedte:Union[float, int]=0,
             naam:str=None):
        """Plot verticale kolommen door middel van coordinaten (positie, hoogte)."""

        self._check_alleen_lezen()
        
        kleur = self.volgende_kleur if kleur is None else kleur

        waardes = self._check_coordinaten(coordinaten)
        X = [x for x, _ in waardes]
        hoogtes = [h for _, h in waardes]
        self.ax.bar(X,
                    height=hoogtes,
                    color=kleur,
                    width=breedte,
                    linewidth=lijn_breedte,
                    edgecolor=lijn_kleur,
                    label=naam)
        return self

    def fx(self,
           functie,
           x:Union[list, tuple],
           breedte=1,
           kleur=None,
           naam:str=None,
           aantal_punten=100):
        """Plot een wiskundige functie bij bepaald domein."""

        self._check_alleen_lezen()
        
        kleur = self.volgende_kleur if kleur is None else kleur

        if len(x) != 2:
            raise ValueError('x waarde moet een lijst of tupel zijn met 2 getallen')

        X = np.linspace(x[0], x[1], aantal_punten)
        f = np.vectorize(functie)
        Y = f(X)

        self.min_x = min(self.min_x, x[0]) if self.min_x is not None else x[0]
        self.max_x = max(self.max_x, x[1]) if self.max_x is not None else x[1]
        self.min_y = min(self.min_y, min(Y)) if self.min_y is not None else min(Y)
        self.max_y = max(self.max_y, max(Y)) if self.max_y is not None else max(Y)

        self.ax.plot(X, Y, label=naam, color=kleur, linewidth=breedte)
        return self

    def _afronden_figuur(self):
        """Rond figuur af."""
        if not self.alleen_lezen:
            min_x = self.min_x if self.min_x is not None else 0
            max_x = self.max_x if self.max_x is not None else 1
            min_y = self.min_y if self.min_y is not None else 0
            max_y = self.max_y if self.max_y is not None else 1

            marge_x = (max_x - min_x) * self.marge_x
            marge_y = (max_y - min_y) * self.marge_y

            self.ax.axis([min_x - marge_x, max_x + marge_x, min_y - marge_y, max_y + marge_y])

            if self.maak_raster:
               self.ax.grid(color='grey', linestyle='-', linewidth=0.2)

            if self.maak_legenda:
                self.ax.legend()

        self.alleen_lezen = True

    def plot(self):
        """Rond figuur af en laat deze inline zien."""
        self._afronden_figuur()
        self.fig.show()
        return self

    def plot_venster(self):
        """Rond figuur af en laat deze in venster zien."""
        self._afronden_figuur()

        pyco.FiguurVenster(
            figuur=self,
            breedte=800,
            hoogte=600,
            titel='figuur' if self.titel == '' else self.titel,
        )

        plt.close(fig=self.fig)
        return self
    
    @property
    def png_url(self):
        """Genereer PNG data encoded url."""
        buf = BytesIO()
        self.fig.savefig(buf, format='png')
        data = base64.b64encode(buf.getbuffer()).decode('ascii')
        url = (f"data:image/png;base64,{data}")
        return url

    @property
    def png_html(self):
        """Genereer PNG code voor inline gebruik IMG HTML."""
        url = self.png_url
        return f"<img src='{url}'/>"

    def bewaar_als_png(self):
        """Vraag om bestandsnaam en bewaar afbeelding als PNG bestand."""
        bestandsnaam = pyco.BestandsnaamVenster(
            extensie='png',
            titel='Bewaren als PNG',
        ).bestandsnaam

        if bestandsnaam is not None:
            self.fig.savefig(bestandsnaam, format='png')

    @property
    def svg_html(self):
        """Genereer SVG code voor inline gebruik HTML."""
        buf = StringIO()
        self.fig.savefig(buf, format='svg')
        data = buf.getvalue()
        # data heeft XML en DOCTYPE header; deze er afhalen (alleen SVG tags)
        data = '<svg ' + data.split('<svg ', 1)[1]
        return data

    def bewaar_als_svg(self):
        """Vraag om bestandsnaam en bewaar afbeelding als SVG bestand."""
        bestandsnaam = pyco.BestandsnaamVenster(
            extensie='svg',
            titel='Bewaren als SVG',
        ).bestandsnaam

        if bestandsnaam is not None:
            self.fig.savefig(bestandsnaam, format='svg')



# afhankelijke externe bibliotheken
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


###############################################################################


import pyco.basis
BasisObject = pyco.basis.BasisObject

import pyco.waarde
Waarde = pyco.waarde.Waarde

import pyco.lijst
Lijst = pyco.lijst.Lijst

import pyco.data
Data = pyco.data.Data

import pyco.knoop
Knoop = pyco.knoop.Knoop

import pyco.lijn
Lijn = pyco.lijn.Lijn

import pyco.vorm
Vorm = pyco.vorm.Vorm

if True:
    import pyco.rechthoek
    Rechthoek = pyco.rechthoek.Rechthoek

    import pyco.cirkel
    Cirkel = pyco.cirkel.Cirkel
    
import pyco.venster
TekstVenster = pyco.venster.TekstVenster
FiguurVenster = pyco.venster.FiguurVenster
BestandsnaamVenster = pyco.venster.BestandsnaamVenster

import pyco.figuur
Figuur = pyco.figuur.Figuur

import pyco.materiaal
Materiaal = pyco.materiaal.Materiaal


###############################################################################


def functies_print_help():
    print("""
    
+-------------------------------------------+
|  algemene pyco functies en eigenschappen  |
+-------------------------------------------+

ALGEMEEN GEBRUIK VAN FUNCTIES         alle namen met () erachter zijn functies
    pc.wortel(9) == 3.0               direct aan te roepen vanuit pc object
    
ALGEMEEN GEBRUIK VAN EIGENSCHAPPEN
    pc.pi == 3.141592653589793        direct aan te roepen vanuit pc object

WISKUNDIGE FUNCTIES                   (geïmporteerd uit Numpy module)
    invoerwaarden:  int, float, np.array, Waarde of Lijst
    sin(x)                            sinus
    cos(x)                            cosinus
    tan(x)                            tangens
    asin(x)                           arcsinus (omgekeerde sin)
    acos(x)                           arccosinus (omgekeerde cos)
    atan(x)                           arctangens (omgekeerde tan)
    hypot(a, b)                       hypotenuse (c in: a^2 + b^c = c^2)
    graden(rad)                       van radialen naar graden
    radialen(deg)                     van graden naar radialen
    sinh(x)                           hyperbolische sinus
    cosh(x)                           hyperbolische cosinus
    tanh(x)                           hyperbolische tangens
    asinh(x)                          arc hyperb. sinus (omgekeerde sinh)
    acosh(x)                          arc hyperb. cosinus (omgekeerde cosh)
    atanh(x)                          arc hyperb. tangens (omgekeerde tanh)
    afronden(x, n)                    rond af op n decimalen (standaard 0)
    plafond(x)                        rond af naar boven (geheel getal)
    vloer(x)                          rond af naar beneden (geheel getal)
    plafond_0_vloer(x)                rond af richting 0 (geheel getal)
    som(lijst)                        de som van de elementen
    product(lijst)                    het product van de elementen
    verschil(lijst)                   lijst met verschillen tussen elementen
    optellen(a, b)                    a + b
    aftrekken(a, b)                   a - b
    vermenigvuldigen(a, b)            a * b
    delen(a, b)                       a / b
    delen_aantal(a, b)                delen afgerond naar beneden
    delen_rest(a, b)                  restant na afronden naar beneden
    macht(a, n)                       a ** n
    reciproke(x)                      1 / x
    negatief(x)                       -x
    kruisproduct(lijst_a, lijst_b)    a x b: staat loodrecht op vector a en b
    inwendigproduct(lijst_a, lijst_b) a . b: is |a| * |b| * cos(theta)
    exp(x)                            exponentieel: berekent e^x
    ln(x)                             natuurlijke logaritme (grondgetal e)
    log(x)                            logaritme met grondgetal 10
    kleinste_gemene_veelvoud(a, b)    kleinste gemene veelvoud: a=12 b=20: 60
    grootste_gemene_deler(a, b)       grootste gemene deler: a=12 b=20: 4
    min(lijst)                        bepaalt minimum waarde lijst
    max(lijst)                        bepaalt maximum waarde lijst
    bijsnijden(lijst, min, max)       snij alle elementen af tot minmax bereik
    wortel(x)                         vierkantswortel
    wortel3(x)                        kubieke wortel
    absoluut(x)                       absolute waarde (altijd positief)
    teken(x)                          positief getal: 1.0   negatief: -1.0 
    kopieer_teken(a, b)               neem getal a, met het teken (+-) van b
    is_positief(a, b)                 stap functie:a<0 -> 0, a=0 -> b, a>0 -> 1 
    verwijder_nan(lijst)              verwijder niet-getallen (not a number)
    getal_nan_inf(lijst)              vervang: nan=0, inf=1.7e+308 (heel groot)
    interpoleer(x, array_x, array_y)  interpoleer x in y; array_x MOET oplopen
    van_totmet_n(van, tot_met, n)     genereert vast aantal getallen (incl. t/m)
    van_tot_stap(van, tot, stap)      genereert vaste stappen (excl. tot)
    gemiddelde(lijst)                 bepaalt het gemiddelde
    stdafw_pop(lijst)                 bepaalt standaardafwijking voor populatie
    stdafw_n(lijst)                   bepaalt standaardafwijking steekproef
    mediaan(lijst)                    bepaalt de mediaan
    percentiel(lijst, percentage)     percentage getal tussen 0 en 100
    correlatie(lijst_a, lijst_b)      bepaalt correlatie matrix
    sorteer(lijst)                    sorteert een lijst van klein naar groot
    omdraaien(lijst)                  draai de volgorde van de lijst om
    is_nan(x)                         bepaalt of waarde een niet-getal is
    is_inf(x)                         bepaalt of waarde oneindig is
    gelijk(lijst_a, lijst_b)          per element kijken of waarden gelijk zijn
    niet_gelijk(lijst_a, lijst_b)     per element kijken of waarden gelijk verschillen
    groter(lijst_a, lijst_b)          per element kijken of waarde groter dan
    groter_gelijk(lijst_a, lijst_b)   idem, maar dan ook gelijk
    kleiner(lijst_a, lijst_b)         per element kijken of waarde kleiner dan
    kleiner_gelijk(lijst_a, lijst_b)  idem, maar dan ook gelijk
    alle(lijst)                       kijkt of alle elementen True zijn
    sommige(lijst)                    kijkt of er minimaal 1 element True is
    niet_alle(lijst)                  kijkt of er minimaal 1 element False is
    geen(lijst)                       kijkt of alle elementen False zijn
    of(a, b)                          kijkt of a of b True is
    en(a, b)                          kijkt of a en b True is
    niet(x)                           omdraaien van True naar False en andersom
    xof(a, b)                         True als a of b True is, en niet beide
    
WISKUNDIGE EIGENSCHAPPEN              (gebaseerd op Numpy module)
    nan                               float die geen getal is (not a number)
    inf                               oneindig groot
    pi                                3.141592653589793
    e                                 2.718281828459045

    """.strip())
    
    
###############################################################################

# eigenschappen

nan = np.nan
inf = np.inf
pi = np.pi
e = np.e

    
###############################################################################
    
# functies

_numpy_functions = dict(
    # tuple(n_args, fn, check_type_eenheid, pre_verander_eenheid_fn, post_verander_eenheid_fn, maak_bool)
    sin = (1, np.sin, 'deg', lambda eh, _: 'rad', lambda eh, _: None),
    cos = (1, np.cos, 'deg', lambda eh, _: 'rad', lambda eh, _: None),
    tan = (1, np.tan, 'deg', lambda eh, _: 'rad', lambda eh, _: None),
    asin = (1, np.arcsin, '-', None, lambda eh, _: 'rad'),
    acos = (1, np.arccos, '-', None, lambda eh, _: 'rad'),
    atan = (1, np.arctan, '-', None, lambda eh, _: 'rad'),
    hypot = (2, np.hypot), 
    graden = (1, np.degrees, '-', None, None),
    radialen = (1, np.radians, '-', None, None),
    sinh = (1, np.sinh, 'deg', lambda eh, _: 'rad', lambda eh, _: None),
    cosh = (1, np.cosh, 'deg', lambda eh, _: 'rad', lambda eh, _: None),
    tanh = (1, np.tanh, 'deg', lambda eh, _: 'rad', lambda eh, _: None),
    asinh = (1, np.arcsinh, '-', None, lambda eh, _: 'rad'),
    acosh = (1, np.arccosh, '-', None, lambda eh, _: 'rad'),
    atanh = (1, np.arctanh, '-', None, lambda eh, _: 'rad'),
    afronden = (2, np.round),
    plafond = (1, np.ceil),
    vloer = (1, np.floor),
    plafond_0_vloer = (1, np.fix),
    som = (1, np.sum),
    product = (1, np.prod),
    verschil = (1, np.diff),
    optellen = (2, np.add),
    aftrekken = (2, np.subtract),
    vermenigvuldigen = (2, np.multiply, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,           
        lambda eh, args: (
            Waarde(1, (Waarde(1, args[0].eenheid)._eenheidbreuk)*(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid 
            if (isinstance(args[0], Waarde) and isinstance(args[1], Waarde))
                  or (isinstance(args[0], Lijst) and isinstance(args[1], Lijst))
            else '@' + eh if eh is not None else None  # @: eenheid naar origineel terugzetten
        )
    ),
    delen = (2, np.divide, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,           
        lambda eh, args: (# Waarde/Lijst delen door Waarde/Lijst
              Waarde(1, (Waarde(1, args[0].eenheid)._eenheidbreuk)/(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid
              if (isinstance(args[0], Waarde) and isinstance(args[1], Waarde))
                  or (isinstance(args[0], Lijst) and isinstance(args[1], Lijst))
              else (
                  # Waarde/Lijst delen door getal
                  args[0].eenheid
                  if sum(isinstance(args[0], t) for t in (Waarde, Lijst))
                     and not sum(isinstance(args[1], t) for t in (Waarde, Lijst))
                  else (
                      # getal delen door Waarde/Lijst
                      Waarde(1, 1/(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid
                      if not sum(isinstance(args[0], t) for t in (Waarde, Lijst))
                          and sum(isinstance(args[1], t) for t in (Waarde, Lijst))
                      else
                          # getal delen door getal
                          None
             )))
    ),
    delen_aantal = (2, np.floor_divide, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,           
        lambda eh, args: (# Waarde/Lijst delen door Waarde/Lijst
             Waarde(1, (Waarde(1, args[0].eenheid)._eenheidbreuk)/(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid
             if (isinstance(args[0], Waarde) and isinstance(args[1], Waarde))
                 or (isinstance(args[0], Lijst) and isinstance(args[1], Lijst))
             else (
                 # Waarde/Lijst delen door getal
                 args[0].eenheid
                 if sum(isinstance(args[0], t) for t in (Waarde, Lijst))
                    and not sum(isinstance(args[1], t) for t in (Waarde, Lijst))
              else (
                  # getal delen door Waarde/Lijst
                     Waarde(1, 1/(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid
                     if not sum(isinstance(args[0], t) for t in (Waarde, Lijst))
                         and sum(isinstance(args[1], t) for t in (Waarde, Lijst))
                     else
                         # getal delen door getal
                         None
            )))
    ),
    delen_rest = (2, np.remainder, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,           
        lambda eh, args: (# Waarde/Lijst delen door Waarde/Lijst
             Waarde(1, (Waarde(1, args[0].eenheid)._eenheidbreuk)/(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid
             if (isinstance(args[0], Waarde) and isinstance(args[1], Waarde))
                 or (isinstance(args[0], Lijst) and isinstance(args[1], Lijst))
             else (
                 # Waarde/Lijst delen door getal
                 args[0].eenheid
                 if sum(isinstance(args[0], t) for t in (Waarde, Lijst))
                    and not sum(isinstance(args[1], t) for t in (Waarde, Lijst))
              else (
                  # getal delen door Waarde/Lijst
                     Waarde(1, 1/(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid
                     if not sum(isinstance(args[0], t) for t in (Waarde, Lijst))
                         and sum(isinstance(args[1], t) for t in (Waarde, Lijst))
                     else
                         # getal delen door getal
                         None
            )))
    ),
    macht = (2, np.power, None, 
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,
        lambda eh, args: Waarde(1, (Waarde(1, eh)._eenheidbreuk)**args[1], config=None).eenheid
    ),
    reciproke = (1, np.reciprocal, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,
        lambda eh, args: Waarde(1, (1/Waarde(1, eh)._eenheidbreuk), config=None).eenheid),
    negatief = (1, np.negative, None, lambda eh, _: eh, lambda eh, _: eh),
    kruisproduct = (2, np.cross, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,           
        lambda eh, args: (
            Waarde(1, (Waarde(1, args[0].eenheid)._eenheidbreuk)*(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid 
            if (isinstance(args[0], Waarde) and isinstance(args[1], Waarde))
                  or (isinstance(args[0], Lijst) and isinstance(args[1], Lijst))
            else '@' + eh if eh is not None else None  # @: eenheid naar origineel terugzetten
        )
    ),
    inwendigproduct = (2, np.dot, None,
        lambda eh, _: Waarde(1, (Waarde(1, eh)._eenheidbreuk), config=None).eenheid,           
        lambda eh, args: (
            Waarde(1, (Waarde(1, args[0].eenheid)._eenheidbreuk)*(Waarde(1, args[1].eenheid)._eenheidbreuk), config=None).eenheid 
            if (isinstance(args[0], Waarde) and isinstance(args[1], Waarde))
                  or (isinstance(args[0], Lijst) and isinstance(args[1], Lijst))
            else '@' + eh if eh is not None else None  # @: eenheid naar origineel terugzetten
        )
    ),
    exp = (1, np.exp, None, None, lambda eh, _: None),
    ln = (1, np.log, None, None, lambda eh, _: None),
    log = (1, np.log10, None, None, lambda eh, _: None),
    kleinste_gemene_veelvoud = (2, np.lcm, None, None, lambda eh, _: None),
    grootste_gemene_deler = (2, np.gcd, None, None, lambda eh, _: None),
    bijsnijden = (3, np.clip),
    wortel = (1, np.sqrt, None, None, lambda eh, _: None),
    wortel3 = (1, np.cbrt, None, None, lambda eh, _: None),
    absoluut = (1, np.absolute),
    teken = (1, np.sign, None, None, lambda eh, _: None),
    kopieer_teken = (2, np.copysign),
    is_positief = (2, np.heaviside, None, None, lambda eh, _: None),
    getal_nan_inf = (1, np.nan_to_num),
    verwijder_nan = (1, lambda x: x[np.logical_not(np.isnan(x))]),
    interpoleer = (3, np.interp),
    van_totmet_n = (3, np.linspace),
    van_tot_stap = (3, np.arange),
    gemiddelde = (1, np.mean),
    stdafw_pop = (1, lambda x: np.std(x, ddof=0)),
    stdafw_n = (1, lambda x: np.std(x, ddof=1)),
    mediaan = (1, np.median),
    percentiel = (2, np.percentile),
    correlatie = (2, lambda l1, l2: np.corrcoef(l1, l2)[0, 1], None, None, lambda eh, _: None),
    sorteer = (1, np.sort),
    omdraaien = (1, np.flip),
    is_nan = (1, np.isnan, None, None, lambda eh, _: None, True),
    is_inf = (1, np.isinf, None, None, lambda eh, _: None, True),
    gelijk = (2, np.equal, None, None, lambda eh, _: None, True),
    niet_gelijk = (2, np.not_equal, None, None, lambda eh, _: None, True),
    groter = (2, np.greater, None, None, lambda eh, _: None, True),
    groter_gelijk = (2, np.greater_equal, None, None, lambda eh, _: None, True),
    kleiner = (2, np.less, None, None, lambda eh, _: None, True),
    kleiner_gelijk = (2, np.less_equal, None, None, lambda eh, _: None, True),
    alle = (1, np.all, None, None, lambda eh, _: None, True),
    sommige = (1, np.any, None, None, lambda eh, _: None, True),
    niet_alle = (1, lambda x: ~np.all(x), None, None, lambda eh, _: None, True),
    geen = (1, lambda x: ~np.any(x), None, None, lambda eh, _: None, True),
    of = (2, np.logical_or, None, None, lambda eh, _: None, True),
    en = (2, np.logical_and, None, None, lambda eh, _: None, True),
    niet = (1, np.logical_not, None, None, lambda eh, _: None, True),
    xof = (2, np.logical_xor, None, None, lambda eh, _: None, True),
)
_numpy_functions['min'] = np.amin # reserverd keywords
_numpy_functions['max'] = np.amax
_numpy_functions['abs'] = np.fabs

def _wrap_functie(n_args, fn, check_type_eenheid=None, pre_verander_eenheid_fn=None,
                  post_verander_eenheid_fn=None, maak_bool=False):
    def return_fn(*args, **kwargs):
        if n_args is not None:
            if n_args != len(args):
                raise ValueError('Onjuist aantal argumenten gegeven in functie.')
            args = args[0:n_args]
        # pre
        eenheid = None
        originele_eenheid = None
        waarde = False
        nieuwe_args = []
        for iarg, arg in enumerate(args):
            if isinstance(arg, Waarde):
                arg = arg.kopie()
                if (check_type_eenheid is not None
                        and arg.eenheid is not None
                        and arg._eenheidbreuk != Waarde(1)[check_type_eenheid]._eenheidbreuk):
                    raise ValueError('Eenheid \'{}\' in functie \'{}\' is niet geldig. '
                                     'Dit moet bijvoorbeeld \'{}\' zijn.'.format(
                                     arg.eenheid, fn.__name__, check_type_eenheid))
                waarde = True
                originele_eenheid = (arg.eenheid if originele_eenheid is None
                                                 else originele_eenheid)
                if pre_verander_eenheid_fn is not None:
                    eenheid = arg.eenheid if eenheid is None else eenheid
                    eenheid = pre_verander_eenheid_fn(eenheid, args)
                    arg = arg.gebruik_eenheid(eenheid, check_type=False)
                else:
                    if iarg == 0:
                        eenheid = arg.eenheid
                    else:
                        arg = arg.eh(eenheid)
                nieuwe_args.append(float(arg))
            elif isinstance(arg, Lijst):
                arg = arg.kopie()
                arg_w = Waarde(1)[arg.eenheid]
                if (check_type_eenheid is not None
                        and arg.eenheid is not None
                        and arg_w._eenheidbreuk != Waarde(1)[check_type_eenheid]._eenheidbreuk):
                    raise ValueError('Eenheid \'{}\' in functie \'{}\' is niet geldig. '
                                     'Dit moet bijvoorbeeld \'{}\' zijn.'.format(
                                     arg_w.eenheid, fn.__name__, check_type_eenheid))
                originele_eenheid = arg.eenheid if originele_eenheid is None else originele_eenheid
                if pre_verander_eenheid_fn is not None:
                    eenheid = arg.eenheid if eenheid is None else eenheid
                    eenheid = pre_verander_eenheid_fn(eenheid, args)
                    arg = arg.gebruik_eenheid(eenheid, check_type=False)
                else:
                    if iarg == 0:
                        eenheid = arg.eenheid
                    else:
                        arg = arg.eh(eenheid)
                nieuwe_args.append(arg.array)
            else:
                nieuwe_args.append(arg)
            
        # call
        value = fn(*nieuwe_args, **kwargs)
        
        # post
        gebruik_originele_eenheid = False
        if post_verander_eenheid_fn is not None:
            eenheid = post_verander_eenheid_fn(eenheid, args)
            if eenheid is not None and eenheid[0] == '@':
                eenheid = eenheid[1:]
                gebruik_originele_eenheid = True
        if isinstance(value, bool) or 'numpy.bool' in str(type(value)):
            return value
        elif isinstance(value, type(np.array([]))):
            l = Lijst(np.array(value, dtype='float64'))
            if eenheid is not None:
                if post_verander_eenheid_fn is not None:
                    l = l.gebruik_eenheid(eenheid, check_type=False)
                else:
                    l = l.gebruik_eenheid(eenheid)
            if gebruik_originele_eenheid:
                l = l.gebruik_eenheid(originele_eenheid)
            if len(l) == 1:
                l = Waarde(l[0]).gebruik_eenheid(l.eenheid)
            if maak_bool:
                l = l.waar()
            return l
        elif waarde or eenheid:
            w = Waarde(value)
            if eenheid is not None:
                if post_verander_eenheid_fn is not None:
                    w = w.gebruik_eenheid(eenheid, check_type=False)
                else:
                    w = w.gebruik_eenheid(eenheid)
            if gebruik_originele_eenheid:
                w = w.gebruik_eenheid(originele_eenheid)
            if maak_bool:
                w = bool(w)
            return w
        return value
    return return_fn

for fn_name, fn_arg in _numpy_functions.items():
    if callable(fn_arg):
        setattr(pyco, fn_name, _wrap_functie(None, fn_arg))
    elif len(fn_arg) == 2:
        setattr(pyco, fn_name, _wrap_functie(fn_arg[0], fn_arg[1]))
    elif len(fn_arg) == 3:
        setattr(pyco, fn_name, _wrap_functie(fn_arg[0], fn_arg[1], fn_arg[2]))
    elif len(fn_arg) == 4:
        setattr(pyco, fn_name, _wrap_functie(fn_arg[0], fn_arg[1], fn_arg[2], fn_arg[3]))
    elif len(fn_arg) == 5:
        setattr(pyco, fn_name, _wrap_functie(fn_arg[0], fn_arg[1], fn_arg[2], fn_arg[3], fn_arg[4]))
    elif len(fn_arg) == 6:
        setattr(pyco, fn_name, _wrap_functie(fn_arg[0], fn_arg[1], fn_arg[2], fn_arg[3], fn_arg[4], fn_arg[5]))

    
###############################################################################
    



import pyco.basis
import pyco.waarde

class pc:
    BasisObject = pyco.basis.BasisObject
    Waarde = pyco.waarde.Waarde

    
class Materiaal(pc.BasisObject):
    """Betreft een materiaal met diverse eigenschappen.

    AANMAKEN MATERIAAL
        m = Materiaal(E=Waarde(210, 'GPa'),     # elasticiteitsmodulus
                      G=Waarde(75, 'GPa'),      # glijdings(schuif)modulus
                      v=Waarde(0.4),            # dwarscontractiecoëfficiënt
                                                               (Poisson factor)
                      sm=Waarde(7850, 'kg/m3'), # soortelijke massa (dichtheid)
                      sg=Waarde(78.5, 'kN/m3')) # soortelijk gewicht
        
    RELATIE E, G EN v
        Indien E en v bekend zijn, dan wordt G zelf bepaald: G = E/(2(1+v))
        Indien E en G bekend zijn, dan wordt v zelf bepaald: v = (E/2G)-1
        Alledrie grootheden kunnen handmatig worden overschreven.
        
    OPVRAGEN EN AANPASSEN MATERIAALEIGENSCHAPPEN
        m.E                             # retourneert Waarde object
        m.E = Waarde(180).GPa           # past waarde E aan
    """
    
    DEFAULT_E = pc.Waarde(0, 'N/m2') 
    DEFAULT_G =  pc.Waarde(0, 'N/m2')
    DEFAULT_v = pc.Waarde(0) 
    DEFAULT_sm = pc.Waarde(0, 'kg/m3')
    DEFAULT_sg = pc.Waarde(0, 'N/m3')

    def __init__(self, E=None, G=None, v=None, sm=None, sg=None):
        super().__init__()

        self._E = self.DEFAULT_E 
        self._G = self.DEFAULT_G  
        self._v = self.DEFAULT_v  
        self._sm = self.DEFAULT_sm 
        self._sg = self.DEFAULT_sg
        
        if E is not None:
            self.E = E
        if G is not None:
            self.G = G
        if v is not None:
            self.v = v
        if sm is not None:
            self.sm = sm
        if sg is not None:
            self.sg = sg

    def _check_waarde(self, waarde, zelfde_eenheid):
        if not isinstance(waarde, pc.Waarde):
            raise TypeError('parameter is geen waarde object')
        if not waarde & zelfde_eenheid:
            raise ValueError('type eenheid van waarde is incorrect')

    @property
    def E(self):
        """Elasticiteitsmodulus."""
        return self._E
    @E.setter
    def E(self, waarde):
        self._check_waarde(waarde, self.DEFAULT_E)
        self._E = waarde

    @property
    def G(self):
        """Glijdingsmodulus."""
        if (self._G == self.DEFAULT_G
                and self._E != self.DEFAULT_E
                and self._v != self.DEFAULT_v):
            return self.E / (2 * (pc.Waarde(1) + self.v))
        else:
            return self._G
    @G.setter
    def G(self, waarde):
        self._check_waarde(waarde, self.DEFAULT_G)
        self._G = waarde

    @property
    def v(self):
        """Dwarscontractiecoëfficiënt (Poisson factor)."""
        if (self._v == self.DEFAULT_v
                and self._E != self.DEFAULT_E
                and self._G != self.DEFAULT_G):
            return self.E / (2 * self.G) - pc.Waarde(1)
        else:
            return self._v
    @v.setter
    def v(self, waarde):
        self._check_waarde(waarde, self.DEFAULT_v)
        if float(waarde) >= -1.01 and float(waarde) <= -0.99:
            raise ValueError('een dwarscontractiecoefficient van -1 leidt tot een glijdingsmodulus van oneindig groot; deze waarde is niet mogelijk')
        self._v = waarde

    @property
    def sm(self):
        """Soortelijke massa (dichtheid)."""
        return self._sm
    @sm.setter
    def sm(self, waarde):
        self._check_waarde(waarde, self.DEFAULT_sm)
        self._sm = waarde

    @property
    def sg(self):
        """Soortelijk gewicht."""
        return self._sg
    @sg.setter
    def sg(self, waarde):
        self._check_waarde(waarde, self.DEFAULT_sg)
        self._sg = waarde


if __name__ == '__main__':
    m1 = Materiaal()
    print(m1.E, bool(m1.E))
    m1.E = pc.Waarde(2.1e5, 'MPa')
    m1.v = pc.Waarde(0.2)
    m1.sm = pc.Waarde(700e3, 'kg/m3')
    m1.sg = pc.Waarde(75, 'kN/m3')
    print(m1.E, bool(m1.E))
    print(m1.G)



from typing import Union
import math

#import numpy as np
#import matplotlib.pyplot as plt

import pyco.model as pycom

class Doorsnede(pycom.BasisObject):
    """Betreft een samenstelling van vormen met materiaaleigenschappen.

    AANMAKEN DOORSNEDE
        d1 = Doorsnede(Lijn)            invoeren van vorm objecten

    """

    AFRONDEN_NAAR_NUL = 1e-13

    def __init__(self, onderdelen:dict):
        super().__init__()

    def _float(self, waarde):
        """Zet object om naar een float en rond hele kleine waarden af naar nul."""
        f = float(waarde)
        if f > -1*self.AFRONDEN_NAAR_NUL and f < 1*self.AFRONDEN_NAAR_NUL:
            f = 0.0
        return f


if __name__ == '__main__':
    pass



import pyco.waarde
import pyco.lijst

class pc:
    Waarde = pyco.waarde.Waarde
    Lijst = pyco.lijst.Lijst

class Knoop(pc.Lijst):
    """
    Bevat een lijst van getallen of Waarde (x, y, en/of z) met lengte eenheid.

    AANMAKEN KNOOP              eenheid van 1e component, geldt voor geheel
        k = Knoop([x_waarde, y_waarde, z_waarde])   kan oneindig veel dimensies

    AANPASSEN EENHEID           omzetten van eenheid naar andere eenheid
        k.eenheid               huidige eenheid opvragen (tekst of None)
        k.eenheid = 'N/mm2'     eenheid aanpassen
        k.gebruik_eenheid('m')  zelfde als bovenstaande, retourneert object

    OMZETTEN KNOOP NAAR TEKST   resulteert in nieuw string object
        tekst = str(k)          of automatisch met bijvoorbeeld print(w)
        tekst = format(k,'.2f') format configuratie meegeven voor getal

    MOGELIJKE BEWERKINGEN       resulteert in nieuw Knoop object
        k3 = k1 + k2            knoop optellen bij knoop
        k3 = k1 - k2            knoop aftrekken van knoop
        k2 = n * k1             getal vermenigvuldigen met knoop
        k2 = k1 * n             knoop vermenigvuldigen met getal
        k2 = n / k1             getal delen door knoop
        k2 = k1 / n             knoop delen door getal
        k2 = +k1                behoud teken
        k2 = -k1                verander teken (positief vs. negatief)
        for w in k1:            itereert en geeft float/Waarde object terug
        Waarde/getal = k1.x     retourneert 1e element als Waarde object
        Waarde/getal = k1.y     retourneert 2e element als Waarde object
        Waarde/getal = k1.z     retourneert 3e element als Waarde object
        getal = len(k1)         geeft aantal elementen (dimensies) van knoop

    NUMPY BEWERKINGEN           gebruikt array object
        numpy_array = k1.array  retourneert Numpy array object
                                    (bevat allen getallen, zonder eenheid)
        getal = k1[2]           retourneert getal (zonder eenheid) op index
        numpy_array = k1[1:3]   retourneert Numpy array object vanuit slice

    WAARDEN VERGELIJKEN         resulteert in een boolean (True/False)
        k1 == k2                is gelijk aan
        k1 != k2                is niet gelijk aan
        k1 &  k2                eenheden zijn zelfde type
    """

    def __init__(self, *waardes):
        super().__init__(*waardes)

        if not (self.eenheid is None
                or pc.Waarde(1, self.eenheid) & pc.Waarde(1, 'm')):
            raise ValueError('eenheid van waardes is geen lengte-eenheid')

    @property
    def x(self):
        """Retourneert 1e element."""
        if len(self) > 0:
            return pc.Waarde(self[0], self.eenheid)
        else:
            raise IndexError('knoop heeft geen elementen')

    @property
    def y(self):
        """Retourneert 2e element."""
        if len(self) > 1:
            return pc.Waarde(self[1], self.eenheid)
        else:
            raise IndexError('knoop heeft minder dan twee elementen')

    @property
    def z(self):
        """Retourneert 3e element."""
        if len(self) > 2:
            return pc.Waarde(self[2], self.eenheid)
        else:
            raise IndexError('knoop heeft minder dan drie elementen')



import pickle

class Singleton(type):
    """
    Metaclass om een normale klasse te tronsformeren naar een singleton.
    Ieder nieuwe instantie van klasse wordt genegeerd en er wordt altijd
    de eerste instantie (object) gebruikt.

    class MyClass(BaseClass, metaclass = Singleton):
        pass
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ObjectTeller(metaclass = Singleton):
    """
    ObjectTeller zorgt ervoor dat over hele applicatie objecten een nieuw
    oplopend nummer krijgen. Hierdoor kan achteraf gekeken worden welk object
    eerder was aangemaakt.

    import pyco.basis as pycob

    teller = pycob.ObjectTeller()
    self.object_nummer = teller.object_nummer
    """

    _object_teller = 0

    def __init__(self):
        pass

    @property
    def object_nummer(self):
        self._object_teller += 1
        return self._object_teller

    
class BasisObject(object):
    """Deze klasse dient als onderlegger voor alle klassen."""

    def __init__(self):
        self._documentatie = ''

        # ObjectTeller zorgt ervoor dat over hele applicatie objecten een nieuw
        # oplopend nummer krijgen. Hierdoor kan achteraf gekeken worden welk object
        # eerder was aangemaakt.
        teller = ObjectTeller()
        self._object_nummer = teller.object_nummer
        
    @classmethod
    def van_bestand(cls, pad):
        """Object laden vanuit bestand."""
        bestand = open(pad, 'rb')
        self = pickle.load(bestand)
        bestand.close()
        return self
    
    def naar_bestand(self, pad):
        """Object opslaan als bestand."""
        bestand = open(pad, 'wb')
        pickle.dump(self, bestand)
        bestand.close()

    def __rshift__(self, other:str):
        """Documentatie toevoegen.

        Object() >> "Documentatie."
        """
        self._documentatie = other.strip() if isinstance(other, str) else ''
        return self

    @classmethod
    def print_help(cls):
        name = cls.__name__
        underline = '+--' + len(name)*'-' + '--+'
        docstr = ''
        for line in cls.__doc__.strip().split('\n'):
            line = line[4:] if line[:4] == '    ' else line
            docstr += line + '\n'
        print(f"\n{underline}\n|  {name}  |\n{underline}\n\n{docstr}")



import pandas as pd
import numpy as np

import pyco.basis
import pyco.waarde
import pyco.lijst

class pc:
    BasisObject = pyco.basis.BasisObject
    Waarde = pyco.waarde.Waarde
    Lijst = pyco.lijst.Lijst

class Data(pc.BasisObject):
    """
    Een Pandas DataFrame waarbij eigenschappen een eenheid kunnen hebben.

    AANMAKEN DATA  
        d = Data({'eigenschap1': 'eenheid1',           als eenheid dimensieloos:
                  'eigenschap2': 'eenheid2'})                  '-', '' of None
                  
    EIGENSCHAPPEN
        d.eigenschappen           lijst met aanwezige kolomnamen
        d.eenheden                lijst met eenheden die bij kolommen horen
        d.df                      Pandas DataFrame object     
        
    TOEVOEGEN DATA REGEL    in onderstaande gevallen: 4 eigenschappen (kolommen)
        d.toevoegen([7,5,3,1])             een Python list object
        d.toevoegen(4,5,6,7)               losse argumenten
        d.toevoegen(pc.Lijst(4,9,6,70))    een Lijst object
        d.toevoegen((14,15,16,17))         een Python tuple object

    OPHALEN DATA EIGENSCHAP       
        d['eigenschap1']          een Lijst object
        d.eigenschap1             indien naam eigenschap ook valide Python name
        d[0]                      Python list met waardes van 1e invoer (rij)
        d[3:8]                    Python list met waardes (ook Python list)
                                                    van 4e t/m 8e invoer (rijen)
        d[::2]                    Python list met alle oneven rijnummers
        d[:, 1:3]                 Python list met 2e en 3e kolom
                                                  (alleen waarden, geen eenheid)          
    """

    def __init__(self, lijst_dict):
        super().__init__()
        
        if not isinstance(lijst_dict, dict):
            raise TypeError("type argument 1 is GEEN dict: {}".format(type(lijst_dict)))
            
        lijst_dict = {k:(v if isinstance(v, str) and len(v) > 0 else '-') for k, v in lijst_dict.items()}
        self._dataframe = pd.DataFrame([], columns=pd.MultiIndex.from_tuples(
                [(gh, eh) for gh, eh in lijst_dict.items()]))
        
    @property
    def eigenschappen(self):
        """Lijst met aanwezige kolomnamen."""
        return [es for es, _ in self.df.columns]
    
    @property
    def eenheden(self):
        """Lijst met eenheden die bij eigenschappen (kolommen) horen"""
        return [eh for _, eh in self.df.columns]

        
    def toevoegen(self, *args):
        """Een nieuwe record toevoegen."""
        if len(args) == 1:
            if isinstance(args[0], list) or isinstance(args[0], tuple):
                args = args[0]
            elif isinstance(args[0], pc.Lijst):
                args = args[0].array
            else:
                args = [args[0]]
                
        if len(args) != len(self.eigenschappen):
            raise ValueError('aantal waardes ({}) is niet gelijk aan aantal kolomnamen ({}: {})'.format(len(args), len(self.eigenschappen), ', '.join(["'{}'".format(e) for e in self.eigenschappen])))
    
        tmp_df = pd.DataFrame([args], columns=pd.MultiIndex.from_tuples(
                [(es, eh) for es, eh in zip(self.eigenschappen, self.eenheden)]))
        self._dataframe = pd.concat([self._dataframe, tmp_df], ignore_index=True)
        
    @property    
    def df(self):
        """Een Pandas DataFrame object retourneren."""
        return self._dataframe
    
    def __getitem__(self, eigenschap_bereik):
        """Retourneert een eigenschap als Lijst (tekst invoer) of een aantal rijen van DataFrame (getal/bereik invoer)."""
        if isinstance(eigenschap_bereik, str):
            eigenschap = eigenschap_bereik
            if not eigenschap in self.eigenschappen:
                raise ValueError('Eigenschap \'{}\' is geen geldige eigenschap ({}).'.format(
                        eigenschap,
                        ', '.join(self.eigenschappen)))
            eenheid = self.eenheden[self.eigenschappen.index(eigenschap)]
            return pc.Lijst(self.df[eigenschap][eenheid].values.tolist()).gebruik_eenheid(eenheid)
        else:
            bereik = eigenschap_bereik
            rijen_lijst = self.df.iloc[bereik].values.tolist()
            return rijen_lijst
        
    def __getattr__(self, name):
        """Retourneert een eigenschap (zie __getitem__) ook als attribuut van object."""
        if name in self.eigenschappen:
            return self.__getitem__(name)
    
    def __repr__(self):
        object_str = self.__str__()
        return 'pyco.Data object:\n' + len(object_str.split('\n')[0])*'-' + '\n' + object_str
    
    def __str__(self):
        return str(self.df)





from typing import Union

import pyco.waarde
import pyco.lijn
import pyco.vorm

class pc:
    Waarde = pyco.waarde.Waarde
    Lijn = pyco.lijn.Lijn
    Vorm = pyco.vorm.Vorm

    
class Rechthoek(pc.Vorm):
    """
    Creeert een rechthoekig Vorm object.
    
    AANMAKEN RECHTHOEK
        r = Rechthoek(breedte=Waarde(300).mm, hoogte=Waarde(500).mm)
        
    Verder zijn alle eigenschappen van toepassing als van een Vorm object. 
    """

    def __init__(self,
                 breedte:Union[pc.Waarde, float, int],
                 hoogte:Union[pc.Waarde, float, int]):

        eenheid = None
        if isinstance(breedte, pc.Waarde):
            eenheid = breedte.eenheid
        b = float(breedte)

        if isinstance(hoogte, pc.Waarde):
            hoogte.eenheid = eenheid
        h = float(hoogte)

        lijn = pc.Lijn([
                [-1/2 * b, -1/2 * h],
                [-1/2 * b,  1/2 * h],
                [ 1/2 * b,  1/2 * h],
                [ 1/2 * b, -1/2 * h],
            ])
        lijn.eenheid = eenheid

        super().__init__(lijn)



from typing import Union
import math

import pyco.waarde
import pyco.lijn
import pyco.vorm

class pc:
    Waarde = pyco.waarde.Waarde
    Lijn = pyco.lijn.Lijn
    Vorm = pyco.vorm.Vorm

    
class Cirkel(pc.Vorm):
    """
    Creeert een cirkelvormig Vorm object.
    
    AANMAKEN CIRKEL
        c = Cirkel(straal=Waarde(3).mm)
        
    Vormeigenschappen worden exact bepaald, daar waar een Vorm object een ronde
    rand benadert met kleine rechte lijnen.
    
    Verder zijn alle eigenschappen en methoden van toepassing als van een
    Vorm object.
    """

    def __init__(self,
                 straal:Union[float, int]):

        eenheid = None
        if isinstance(straal, pc.Waarde):
            eenheid = straal.eenheid
        r = float(straal)

        lijn = pc.Lijn([-1*r, 0]).lijn_cirkelboog(
                    middelpunt=(0,0),
                    gradenhoek=360,
                    stappen=200,
                )
        lijn.eenheid = eenheid

        super().__init__(lijn)

        # corrigeer benadering cirkel met polygoon door exacte waarde
        self.O = 2 * math.pi * r
        self.A = math.pi * r**2
        self.xmin = -r
        self.xmax = r
        self.ymin = -r
        self.ymax = r
        self.Ixx = 1/4 * math.pi * r**4
        self.Iyy = self.Ixx
        self.Ixy = 0.0
        self.I1 = self.Ixx
        self.I2 = self.Ixx
        self.alpha = 0.0
        self.Wxmin = self.Ixx / r
        self.Wxmax = self.Wxmin
        self.Wymin = self.Wxmin
        self.Wymax = self.Wxmin
        self.kxmin = -r / 4
        self.kxmax = r / 4
        self.kymin = self.kxmin
        self.kymax = self.kxmax

        self._bereken_waardes()
